from __future__ import annotations
import numpy as np
import pandas as pd
from autotrader.brokers.trading import Order, Trade, Position
from autotrader.brokers.broker_utils import BrokerUtils


class Broker:
    """Autotrader virtual broker to simulate trading in backtest.

    Attributes
    ----------
    orders : dict
        A dictionary containing Orders submitted.
    trades : dict
        A dictionary containing all trades.
    leverage : int
        The account leverage.
    spread : float
        The average spread to use when opening and closing trades.
    margin_available : float
        The margin available on the account.
    portfolio_balance : float
        The account balance.
    hedging : bool
        Flag whethere hedging is enabled on the account. The default is False.
    home_currency : str
        The default is 'AUD'.
    NAV : float
        The net asset value of the account.
    unrealised_PL : float
        The unrealised (floating) PnL.
    verbosity : int
        The verbosity of the broker.
    commission_scheme : str
        The default is 'percentage'.
    commission : float
        The commission value.

    """
    
    def __init__(self, broker_config: dict, utils: BrokerUtils) -> None:
        self.utils = utils
        
        # Orders
        self.orders = {}
        
        # Trades
        self.trades = {}
        
        # Account 
        self.leverage = 1
        self.spread = 0 # TODO - pips or price units? Add docs
        self.margin_available = 0
        self.portfolio_balance = 0
        self.hedging = False
        self.margin_closeout = 0.0 # Fraction at margin call
        
        self.profitable_trades = 0
        self.peak_value = 0
        self.low_value = 0
        self.max_drawdown = 0
        self.home_currency = 'AUD'
        self.NAV = 0
        self.unrealised_PL = 0
        self.verbosity = broker_config['verbosity']
        
        # Commissions
        self.commission_scheme = 'percentage'
        self.commission = 0
        
    
    def __repr__(self):
        return 'AutoTrader Virtual Broker'
    
    
    def __str__(self):
        return 'AutoTrader Virtual Broker'
    
    
    def get_NAV(self) -> float:
        """Returns Net Asset Value of account.
        """
        return self.NAV
    
    
    def get_balance(self) -> float:
        """Returns balance of account.
        """
        return self.portfolio_balance
    
    
    def place_order(self, order: Order, **kwargs) -> None:
        """Place order with broker.
        """
        
        # Call order to set order time
        datetime_stamp = kwargs['order_time']
        order(order_time = datetime_stamp)
        
        if order.order_type == 'limit' or order.order_type == 'stop-limit':
            working_price = order.order_limit_price
        else:
            working_price = order.order_price
        
        # Convert stop distance to price
        if not order.stop_loss and order.stop_distance:
            pip_value = self.utils.get_pip_ratio(order.instrument)
            order.stop_loss = working_price - order.direction*order.stop_distance*pip_value
        
        # Verify SL price
        invalid_order = False
        if order.stop_loss and order.direction*(working_price - order.stop_loss) < 0:
            direction = 'long' if order.direction > 0 else 'short'
            SL_placement = 'below' if order.direction > 0 else 'above'
            reason = "Invalid stop loss request: stop loss must be "+ \
                    f"{SL_placement} the order price for a {direction}" + \
                    " trade order.\n"+ \
                    f"Order Price: {working_price}\nStop Loss: {order.stop_loss}"
            invalid_order = True
        
        # Verify TP price
        if order.take_profit is not None and order.direction*(working_price - order.take_profit) > 0:
            direction = 'long' if order.direction > 0 else 'short'
            TP_placement = 'above' if order.direction > 0 else 'below'
            reason = "Invalid take profit request: take profit must be "+ \
                  f"{TP_placement} the order price for a {direction}" + \
                  " trade order.\n"+ \
                  f"Order Price: {working_price}\nTake Profit: {order.take_profit}"
            invalid_order = True
        
        # Append order to orders
        order.id = self._get_new_order_id()
        self.orders[order.id] = order
        
        # Submit order
        if invalid_order:
            if self.verbosity > 0:
                print(f"  Order {order.id} rejected.\n")
            self.cancel_order(order.id, reason)
        else:
            immediate_orders = ['close', 'reduce']
            status = 'open' if order.order_type in immediate_orders else 'pending'
            order.status = status
        
    
    def get_orders(self, instrument: str = None, 
                   order_status: str = 'open') -> dict:
        """Returns orders.
        """
        orders = {}
        if instrument:
            for order_id, order in self.orders.items():
                if order.instrument == instrument and order.status == order_status:
                    orders[order_id] = order
        else:
            for order_id, order in self.orders:
                if order.status == order_status:
                    orders[order_id] = order
        return orders
    
    
    def cancel_order(self, order_id: int, reason: str = None) -> None:
        """Cancels the order.
        """
        self.orders[order_id].reason = reason
        self.orders[order_id].status = 'cancelled'
        
        if self.verbosity > 0 and reason:
            print(reason)
    
    
    def get_trades(self, instruments: str | list = None,
                   trade_status: str = 'open') -> dict:
        """Returns open trades for the specified instrument.
        """
        open_trades = {}
        if instruments:
            # Specific instruments requested
            if type(instruments) is str:
                # Single instrument provided, put into list
                instruments = [instruments]
                
            for trade_id, trade in self.trades.items():
                if trade.instrument in instruments and trade.status == trade_status:
                    open_trades[trade_id] = trade
        else:
            # Return all currently open positions
            for trade_id, trade in self.trades.items():
                if trade.status == trade_status:
                    open_trades[trade_id] = trade
        
        return open_trades
    
    
    def get_trade_details(self, trade_ID: int) -> Trade:
        """Returns the trade specified by trade_ID.
        """
        return self.trades[trade_ID]
    
    
    def get_positions(self, instruments: str = None) -> dict:
        """Returns the open positions (including all open trades) in the account.
        """
        if instruments:
            # instruments provided, check type
            if type(instruments) is str:
                # Single instrument provided, put into list
                instruments = [instruments]
        else:
            # No specific instrument requested, get all open
            instruments = []
            for trade_id, trade in self.trades.items(): 
                instruments.append(trade.instrument)
            
        open_positions = {}
        for instrument in instruments:
            # First get open trades
            open_trades = self.get_trades(instrument)
            
            if len(open_trades) > 0:
                # Trades exist for current instrument, collate
                long_units = 0
                long_PL = 0
                long_margin = 0
                short_units = 0
                short_PL = 0
                short_margin = 0
                total_margin = 0
                trade_IDs = []
                
                for trade_id in open_trades:
                    trade_IDs.append(trade_id)
                    total_margin += open_trades[trade_id].margin_required
                    if open_trades[trade_id].direction > 0:
                        # Long trade
                        long_units += open_trades[trade_id].size
                        long_PL += open_trades[trade_id].unrealised_PL
                        long_margin += open_trades[trade_id].margin_required
                    else:
                        # Short trade
                        short_units += open_trades[trade_id].size
                        short_PL += open_trades[trade_id].unrealised_PL
                        short_margin += open_trades[trade_id].margin_required
            
                # Construct instrument position dict
                instrument_position = {'long_units': long_units,
                                       'long_PL': long_PL,
                                       'long_margin': long_margin,
                                       'short_units': short_units,
                                       'short_PL': short_PL,
                                       'short_margin': short_margin,
                                       'total_margin': total_margin,
                                       'trade_IDs': trade_IDs,
                                       'instrument': instrument,
                                       'net_position': long_units-short_units}
                
                instrument_position = Position(**instrument_position)
                
                # Append position dict to open_positions dict
                open_positions[instrument] = instrument_position
                
        return open_positions
    
    
    def get_margin_available(self) -> float:
        """Returns the margin available on the account.
        """
        return self.margin_available
    
    
    def _fill_order(self, order_id: int, candle: pd.core.series.Series, 
                    limit_price: float = None) -> None:
        """Fills an open order.
        
        Notes
        -----
        If hedging is not enabled, any orders which are contrary to an open
        position will first reduce (or close) the open position before 
        being filled. If the units remaining of the order after reducing the 
        open position exceed margin requirements, the entire order will be 
        cancelled, and the original position will not be impacted.
        """
        
        order = self.orders[order_id]
        
        close_existing_position = False
        if not self.hedging:
            current_position = self.get_positions(order.instrument)
            net_position = current_position[order.instrument].net_position \
                if current_position else order.direction
            if order.direction != np.sign(net_position):
                if order.size > abs(net_position):
                    # Modify order size to remaining units
                    order.size -= abs(net_position)
                    close_existing_position = True
                    
                else:
                    # Reduce the current position
                    self._reduce_position(order)
                    order.status = 'closed'
                    return
            
        # Calculate margin requirements
        current_price = candle.Open
        pip_value = self.utils.get_pip_ratio(order.instrument) # TODO - implications on non FX?
        position_value = order.size * current_price * order.HCF
        margin_required = self._calculate_margin(position_value)
        spread_cost = order.size * self.spread * pip_value
        spread_shift = 0.5 * order.direction * self.spread * pip_value
        
        if margin_required < self.margin_available:
            # Determine working price
            if limit_price is None:
                working_price = candle.Open + spread_shift
            else:
                working_price = limit_price + spread_shift
                
            if close_existing_position:
                # Close the open position before proceeding
                self._close_position(order.instrument, candle, working_price)
            
            # Fill order
            trade_id = self._get_new_trade_id()
            trade = Trade(order)
            trade.id = trade_id
            trade.fill_price = working_price
            trade.time_filled = candle.name
            trade.margin_required = margin_required
            self.trades[trade_id] = trade
            
            # Subtract spread cost from account NAV
            self.NAV -= spread_cost
            
        else:
            # Cancel order
            cancel_reason = "Insufficient margin to fill order."
            self.cancel_order(order_id, cancel_reason)
    
    
    def _update_positions(self, candle: pd.core.series.Series, 
                          instrument: str) -> None:
        """Updates orders and open positions based on current candle.
        """
        # Open pending orders
        pending_orders = self.get_orders(instrument, 'pending')
        for order_id, order in pending_orders.items():
            if candle.name > order.order_time:
                order.status = 'open'
        
        # Update open orders
        for order_id, order in self.orders.items():
            # Filter orders by instrument type since candle is instrument specific
            if order.instrument == instrument and order.status == 'open':
                if order.order_type == 'market':
                    # Market order type - proceed to fill
                    self._fill_order(order_id, candle)
                
                elif order.order_type == 'stop-limit':
                    # Check if order_stop_price has been reached yet
                    if candle.Low < order.order_stop_price < candle.High:
                        # order_stop_price has been reached, change order type to 'limit'
                        order.order_type = 'limit'
                
                elif order.order_type == 'modify':
                    # Modification order
                    self._modify_order(order)
                
                elif order.order_type == 'close':
                    related_order = order.related_orders
                    self._close_position(order.instrument, candle, 
                                         candle.Close, trade_id = related_order)
                    order.status = 'closed'
                elif order.order_type == 'reduce':
                    self._reduce_position(order)
                    order.status = 'closed'
                    
                # Check for limit orders
                if order.order_type == 'limit':
                    # Limit order type
                    if order.direction > 0:
                        if candle.Low < order.order_limit_price:
                            self._fill_order(order_id, candle, 
                                             order.order_limit_price)
                    else:
                        if candle.High > order.order_limit_price:
                            self._fill_order(order_id, candle, 
                                             order.order_limit_price)
                
        # Update open trades
        unrealised_PL = 0 # Un-leveraged value
        for trade_id, trade in self.trades.items():
            if trade.instrument == instrument and trade.status == 'open':
                # Update stop losses
                if trade.stop_type == 'trailing':
                    # Trailing stop loss type, check if price has moved SL
                    if trade.stop_distance is not None:
                        pip_value = self.utils.get_pip_ratio(trade.instrument)
                        pip_distance = trade.stop_distance
                        distance = pip_distance*pip_value # price units
                    else:
                        distance = abs(trade.fill_price - \
                                       trade.stop_loss)
                        pip_value = self.utils.get_pip_ratio(trade.instrument)
                        trade.stop_distance = distance / pip_value
                        
                    if trade.direction > 0:
                        # long position, stop loss only moves up
                        new_stop = candle.High - distance
                        if new_stop > trade.stop_loss:
                            self._update_stop_loss(trade_id, new_stop, 
                                                   new_stop_type='trailing')
                    else:
                        # short position, stop loss only moves down
                        new_stop = candle.Low + distance
                        if new_stop < trade.stop_loss:
                            self._update_stop_loss(trade_id, new_stop, 
                                                   new_stop_type='trailing')
                
                # Update trades
                if trade.direction > 0:
                    # Long trade
                    if trade.stop_loss and \
                        candle.Low < trade.stop_loss:
                        # Stop loss hit
                        self._close_position(trade.instrument, 
                                            candle, 
                                            trade.stop_loss,
                                            trade_id)
                    elif trade.take_profit and \
                        candle.High > trade.take_profit:
                        # Take Profit hit
                        self._close_position(trade.instrument, 
                                            candle, 
                                            trade.take_profit,
                                            trade_id)
                    else:
                        # Position is still open, update value of holding
                        size        = trade.size
                        direction   = trade.direction
                        fill_price = trade.fill_price
                        price       = candle.Close
                        HCF         = trade.HCF
                        trade_PL    = direction*size*(price - fill_price)*HCF
                        unrealised_PL += trade_PL
                        
                        # Update PL of trade
                        trade.last_price = price
                        trade.last_time = candle.name
                        trade.unrealised_PL = trade_PL
                
                else:
                    # Short trade
                    if trade.stop_loss is not None and \
                        candle.High > trade.stop_loss:
                        # Stop loss hit
                        self._close_position(trade.instrument, candle, 
                                            trade.stop_loss, trade_id)
                    elif trade.take_profit is not None and \
                        candle.Low < trade.take_profit:
                        # Take Profit hit
                        self._close_position(trade.instrument, candle, 
                                             trade.take_profit, trade_id)
                    else:
                        # Position is still open, update value of holding
                        size        = trade.size
                        direction   = trade.direction
                        fill_price = trade.fill_price
                        price       = candle.Close
                        HCF         = trade.HCF
                        trade_PL    = direction*size*(price - fill_price)*HCF
                        unrealised_PL += trade_PL
                        
                        # Update PL of trade
                        trade.last_price = price
                        trade.last_time = candle.name
                        trade.unrealised_PL = trade_PL
        
        # Update margin available
        self._update_margin(instrument, candle)
        
        # Update unrealised P/L
        self.unrealised_PL = unrealised_PL
        
        # Update open position value
        self.NAV = self.portfolio_balance + self.unrealised_PL
    
    
    def _close_position(self, instrument: str, candle: pd.core.series.Series, 
                        exit_price: float, trade_id=None) -> None:
        """Closes position in instrument.
        """
        if trade_id:
            # single trade specified to close
            self._close_trade(trade_id=trade_id, exit_price=exit_price,
                              candle=candle)
        else:
            # Close all positions for instrument
            for trade_id, trade in self.trades.items():
                if trade.instrument == instrument and trade.status == 'open':
                    self._close_trade(trade_id=trade_id, exit_price=exit_price,
                                      candle=candle)
    
    
    def _close_trade(self, trade_id: int = None, exit_price: float = None,
                     candle: pd.core.series.Series = None) -> None:
        """Closes trade by order number.

        Parameters
        ----------
        trade_id : int, optional
            The trade id. The default is None.
        exit_price : float, optional
            The trade exit price. The default is None.
        candle : pd.core.series.Series, optional
            Slice of OHLC data. The default is None.

        Returns
        -------
        None
            The trade will be marked as closed.
        """
        trade = self.trades[trade_id]
        fill_price = trade.fill_price
        size = trade.size
        direction = trade.direction
        HCF = trade.HCF
        
        # Account for missing inputs
        exit_price = trade.last_price if not exit_price else exit_price
        
        # Update portfolio with profit/loss
        gross_PL = direction*size*(exit_price - fill_price)*HCF
        commission = self._calculate_commissions(trade_id, exit_price, size)
        net_profit = gross_PL - commission
        
        if net_profit > 0:
            self.profitable_trades += 1
        
        # Add trade to closed positions
        trade.profit = net_profit
        trade.balance = self.portfolio_balance
        trade.exit_price = exit_price
        trade.fees = commission
        if candle is None:
            trade.exit_time = trade.last_time
        else:
            trade.exit_time = candle.name 
        trade.status = 'closed'
        
        # Update account
        self._add_funds(net_profit)
        self._update_MDD()
    
    
    def _reduce_position(self, order: Order) -> None:
        """Reduces the position of the specified instrument by FIFO. 
        
        The direction parameter of the order is used to specify whether 
        to reduce long units, or to reduce short units. For example, the
        order details below will reduce long units of the position being 
        held.
        
        """
        # Consired long vs. short units to be reduced
        instrument = order.instrument
        reduction_size = order.size
        reduction_direction = order.direction
        
        # Get open trades for instrument
        open_trades = self.get_trades(instrument)
        # open_trade_IDs = list(open_trades.keys())
        
        # Modify existing trades until there are no more units to reduce
        units_to_reduce = reduction_size
        while units_to_reduce > 0:
            for trade_id, trade in open_trades.items():
                if trade.direction != reduction_direction:
                    # Only reduce long trades when reduction direction is -1
                    # Only reduce short trades when reduction direction is 1
                    if units_to_reduce >= trade.size:
                        # Entire trade must be closed
                        last_price = trade.last_price
                        self._close_trade(trade_id=trade_id, exit_price=last_price)
                        
                        # Update units_to_reduce
                        units_to_reduce -= abs(trade.size)
                        
                    else:
                        # Partially close trade
                        self._partial_trade_close(trade_id, units_to_reduce)
                        
                        # Update units_to_reduce
                        units_to_reduce = 0
                    
    
    def _partial_trade_close(self, trade_id: int, units: float) -> None:
        """Partially closes a trade.
        
        The original trade ID remains, but the trade size may be reduced. The
        portion that gets closed is assigned a new ID.
        """
        trade = self.trades[trade_id]
        
        # Create new trade for reduced amount
        partial_trade = Trade._split(trade, units)
        partial_trade_id = self._get_new_trade_id()
        partial_trade.id = partial_trade_id
        self.trades[partial_trade_id] = partial_trade
        
        # Close partial trade
        self._close_trade(partial_trade_id)
        
    
    def _calculate_commissions(self, trade_id: int, exit_price: float, 
                               units: float = None) -> float:
        """Calculates trade commissions.
        """
        if self.commission_scheme == 'percentage':
            fill_price = self.trades[trade_id].fill_price
            size = self.trades[trade_id].size if units is None else units
            HCF = self.trades[trade_id].HCF
            
            open_trade_value = abs(size)*fill_price*HCF
            close_trade_value = abs(size)*exit_price*HCF
            commission  = (self.commission/100) * (open_trade_value + close_trade_value)
        
        elif self.commission_scheme == 'fixed':
            size = self.trades[trade_id].size if units is None else units
            commission = self.commission * size
            
        return commission
    
    
    def _add_funds(self, amount: float) -> None:
        """Adds funds to brokerage account.
        """
        self.portfolio_balance  += amount
    
    
    def _make_deposit(self, deposit: float) -> None:
        """Adds deposit to account balance and NAV.
        """
        if self.portfolio_balance == 0:
            # If this is the initial deposit, set peak and low values for MDD
            self.peak_value = deposit
            self.low_value = deposit
            
        self.portfolio_balance += deposit
        self.NAV += deposit
        self._update_margin()
    
    
    def _calculate_margin(self, position_value: float) -> float:
        """Calculates margin required to take a position.
        """
        margin = position_value / self.leverage
        return margin
    
    
    def _update_margin(self, instrument: str = None, 
                       candle: pd.core.series.Series = None) -> None:
        """Updates margin available in account.
        """        
        margin_used = 0
        for trade_id, trade in self.trades.items():
            if trade.status == 'open':
                size = trade.size
                HCF = trade.HCF
                last_price = trade.last_price
                position_value = abs(size) * last_price * HCF
                margin_required = self._calculate_margin(position_value)
                margin_used += margin_required
                
                # Update margin required in trade dict
                trade.margin_required = margin_required
        
        self.margin_available = self.NAV - margin_used
        
        # Check for margin call
        if self.margin_available/self.NAV < self.margin_closeout:
            # Margin call
            if self.verbosity > 0:
                print("MARGIN CALL: closing all positions.")
            self._margin_call(instrument, candle)


    def _update_MDD(self) -> None:
        """Function to calculate maximum portfolio drawdown.
        """
        balance     = self.portfolio_balance
        peak_value  = self.peak_value
        low_value   = self.low_value
        
        if balance > peak_value:
            self.peak_value = balance
            self.low_value = balance
            
        elif balance < low_value:
            self.low_value = balance
        
        MDD = 100*(low_value - peak_value)/peak_value
        
        if MDD < self.max_drawdown:
            self.max_drawdown = MDD
    
    
    def _modify_order(self, order: Order) -> None:
        """Modify order with updated parameters. Called when order_type = 'modify', 
        modifies trade specified by related_orders key.
        """
        # Get ID of trade to modify
        modify_trade_id = order.related_orders
        
        if order.stop_loss is not None:
            # New stop loss provided
            self._update_stop_loss(modify_trade_id, order.stop_loss, 
                                   order.stop_type)
            
        if order.take_profit is not None:
            self._update_take_profit(modify_trade_id, order.take_profit)
        
        
    def _update_stop_loss(self, trade_id: int, new_stop_loss: float, 
                          new_stop_type: str = 'limit') -> None:
        """Updates stop loss on open trade.
        """
        self.trades[trade_id].stop_loss = new_stop_loss
        self.trades[trade_id].stop_type = new_stop_type
    
    
    def _update_take_profit(self, trade_id: int, new_take_profit: float) -> None:
        """Updates take profit on open trade.
        """
        self.trades[trade_id].take_profit = new_take_profit
        
        
    def _get_new_order_id(self):
        return len(self.orders) + 1
    
    
    def _get_new_trade_id(self):
        return len(self.trades) + 1
    
    
    def _margin_call(self, instrument: str, candle: pd.core.series.Series):
        """Closes open positions.
        """
        self._close_position(instrument, candle, candle.Close)
    
    


import os
from finta import TA
import autotrader.indicators as indicators
from autotrader.brokers.trading import Order


class SimpleMACD:
    """Simple MACD Strategy
    
    Rules
    ------
    1. Trade in direction of trend, as per 200EMA.
    2. Entry signal on MACD cross below/above zero line.
    3. Set stop loss at recent price swing.
    4. Target 1.5 take profit.
    """
    
    def __init__(self, params, data, instrument):
        """Define all indicators used in the strategy.
        """
        
        self.name = "MACD Trend Strategy"
        self.params = params
        self.instrument = instrument
        
        # Initial feature calculation
        self.calculate_features(data)

        # Construct indicators dict for plotting
        self.indicators = {'MACD (12/26/9)': {'type': 'MACD',
                                              'macd': self.MACD.MACD,
                                              'signal': self.MACD.SIGNAL},
                           'EMA (200)': {'type': 'MA',
                                         'data': self.ema}}
    
    
    def calculate_features(self, data):
        
        # Save data for other functions
        self.data = data
        
        # 200EMA
        self.ema = TA.EMA(data, self.params['ema_period'])
        
        # MACD
        self.MACD = TA.MACD(data, self.params['MACD_fast'], 
                            self.params['MACD_slow'], self.params['MACD_smoothing'])
        self.MACD_CO = indicators.crossover(self.MACD.MACD, self.MACD.SIGNAL)
        self.MACD_CO_vals = indicators.cross_values(self.MACD.MACD, 
                                                    self.MACD.SIGNAL,
                                                    self.MACD_CO)
        
        # Price swings
        self.swings = indicators.find_swings(data)
        
        # TODO - append data to self.indicators as it is created
        
        
    def generate_signal(self, data):
        """Define strategy to determine entry signals.
        """
        # Feature calculation
        self.calculate_features(data)
        i = -1 # Now just use latest data
        if self.data.Close.values[i] > self.ema[i] and \
            self.MACD_CO[i] == 1 and \
            self.MACD_CO_vals[i] < 0:
                exit_dict = self.generate_exit_levels(signal=1, i=i)
                new_order = Order(direction=1,
                                  stop_loss=exit_dict['stop_loss'],
                                  take_profit=exit_dict['take_profit'])
                
        elif self.data.Close.values[i] < self.ema[i] and \
            self.MACD_CO[i] == -1 and \
            self.MACD_CO_vals[i] > 0:
                exit_dict = self.generate_exit_levels(signal=-1, i=i)
                new_order = Order(direction=-1,
                                  stop_loss=exit_dict['stop_loss'],
                                  take_profit=exit_dict['take_profit'])

        else:
            new_order = Order()
        
        return new_order
    
    
    def generate_exit_levels(self, signal, i):
        """Function to determine stop loss and take profit levels.
        """
        stop_type = 'limit'
        RR = self.params['RR']
        
        if signal == 0:
            stop = None
            take = None
        else:
            if signal == 1:
                stop = self.swings.Lows[i]
                take = self.data.Close[i] + RR*(self.data.Close[i] - stop)
            else:
                stop = self.swings.Highs[i]
                take = self.data.Close[i] - RR*(stop - self.data.Close[i])
        
        exit_dict = {'stop_loss': stop, 
                     'stop_type': stop_type,
                     'take_profit': take}
        
        return exit_dict
    
    
if __name__ == "__main__":
    from autotrader.autotrader import AutoTrader
    
    config = {'NAME': 'MACD Strategy',
              'MODULE': 'macd_strategy_continuous',
              'CLASS': 'SimpleMACD',
              'INTERVAL': 'H4',
              'PERIOD': 300,
              'RISK_PC': 1.5,
              'SIZING': 'risk',
              'PARAMETERS': {'ema_period': 200,
                             'MACD_fast': 5,
                             'MACD_slow': 19,
                             'MACD_smoothing': 9,
                             'RR': 1.5},
              'WATCHLIST': ['EUR_USD'],}
    home_dir = os.getcwd()
    
    at = AutoTrader()
    at.configure(verbosity=1, show_plot=False, 
                 mode='continuous', update_interval='4h')
    at.add_strategy(config_dict=config, strategy=SimpleMACD)
    at.plot_settings(show_cancelled=False)
    at.add_data({'EUR_USD': 'EUR_USD_H4.csv'}, 
                data_directory=os.path.join(home_dir, 'data'))
    at.backtest(start = '1/1/2015',
                end = '1/3/2022',
                initial_balance=1000,
                leverage=30,
                spread=0.5,
                commission=0.005)
    at.run()
    
    bot = at.get_bots_deployed()



import v20
import sys
import time
import datetime
import traceback
import numpy as np
import pandas as pd
from autotrader.brokers.broker_utils import BrokerUtils
from autotrader.brokers.trading import Order, Trade, Position


class Broker:
    def __init__(self, oanda_config: dict, utils: BrokerUtils):
        """Create v20 context.
        """
        self.API = oanda_config["API"]
        self.ACCESS_TOKEN = oanda_config["ACCESS_TOKEN"]
        self.port = oanda_config["PORT"]
        self.ACCOUNT_ID = oanda_config["ACCOUNT_ID"]
        self.api = v20.Context(hostname = self.API, 
                               token = self.ACCESS_TOKEN, 
                               port = self.port)
        
        self.STREAM_API = "stream-fxpractice.oanda.com"
        self.stream = v20.Context(hostname = self.STREAM_API, 
                                  token = self.ACCESS_TOKEN, 
                                  port = self.port)
        self.open_positions = {}
        self.utils = utils
        
    
    def __repr__(self):
        return 'AutoTrader-Oanda Broker Interface'
    
    
    def __str__(self):
        return 'AutoTrader-Oanda Broker Interface'
    
    
    def get_NAV(self) -> float:
        """Returns Net Asset Value of account.
        """
        self._check_connection()
        response = self.api.account.get(accountID=self.ACCOUNT_ID)
        return response.body["account"].NAV
    
    
    def get_balance(self) -> float:
        """Returns account balance.
        """
        self._check_connection()
        response = self.api.account.get(accountID=self.ACCOUNT_ID)
        return response.body["account"].balance
    
    
    def place_order(self, order: Order, **kwargs):
        """Submits order to broker.
        """
        self._check_connection()
        
        # Call order to set order time
        order()
        
        # Submit order
        if order.order_type == 'market':
            response = self._place_market_order(order)
        elif order.order_type == 'stop-limit':
            response = self._place_stop_limit_order(order)
        elif order.order_type == 'limit':
            response = self._place_limit_order(order)
        elif order.order_type == 'close':
            response = self._close_position(order.instrument)
        else:
            print("Order type not recognised.")
        
        # Check response
        # output = self._check_response(response)
        
        return response
        
    
    def get_orders(self, instrument=None, **kwargs) -> dict:
        """Get all pending orders in the account. 
        """
        self._check_connection()
        response = self.api.order.list_pending(accountID=self.ACCOUNT_ID, 
                                               instrument=instrument)
        oanda_pending_orders = response.body['orders']
        orders = {}
        
        for order in oanda_pending_orders:
            if order.type != 'TAKE_PROFIT' and order.type != 'STOP_LOSS':
                new_order = {}
                new_order['id'] = order.id
                new_order['status'] = 'open'
                new_order['order_type'] = order.type
                new_order['order_stop_price'] = order.price
                new_order['order_limit_price'] = order.price
                new_order['direction'] = np.sign(order.units)
                new_order['order_time'] = order.createTime
                new_order['instrument'] = order.instrument
                new_order['size'] = abs(order.units)
                new_order['order_price'] = order.price
                
                # Check for take profit
                if order.takeProfitOnFill is not None:
                    new_order['take_profit'] = order.takeProfitOnFill.price
                
                # Check for stop loss
                if order.stopLossOnFill is not None:
                    new_order['stop_loss'] = order.stopLossOnFill.price
                    new_order['stop_type'] = 'limit'
                    
                if instrument is not None and order.instrument == instrument:
                    orders[order.id] = Order._from_dict(new_order)
                elif instrument is None:
                    orders[order.id] = Order._from_dict(new_order)
            
        return orders
    
    
    def cancel_order(self, order_id: int, **kwargs) -> None:
        """Cancels pending order by ID.
        """
        self._check_connection()
        self.api.order.cancel(accountID = self.ACCOUNT_ID, 
                              orderSpecifier=str(order_id))
    
    
    def get_trades(self, instruments=None, **kwargs) -> dict:
        """Returns the open trades held by the account. 
        """
        self._check_connection()
        response = self.api.trade.list_open(accountID=self.ACCOUNT_ID)
        oanda_open_trades = response.body['trades']
        
        open_trades = {}
        for trade in oanda_open_trades:
            new_trade = {}
            related_orders = []
            new_trade['order_ID']           = trade.id
            new_trade['order_stop_price']   = trade.price
            new_trade['order_limit_price']  = trade.price
            new_trade['direction']          = np.sign(trade.currentUnits)
            new_trade['order_time']         = trade.openTime
            new_trade['instrument']         = trade.instrument
            new_trade['size']               = abs(trade.currentUnits)
            new_trade['order_price']        = trade.price
            new_trade['entry_price']        = trade.price
            
            # Check for take profit
            if trade.takeProfitOrder is not None:
                new_trade['take_profit']    = trade.takeProfitOrder.price
                related_orders.append(trade.takeProfitOrder.id)
            
            # Check for stop loss
            if trade.stopLossOrder is not None:
                new_trade['stop_loss']    = trade.stopLossOrder.price
                new_trade['stop_type']    = 'limit'
                related_orders.append(trade.stopLossOrder.id)
            
            if related_orders is not None:
                new_trade['related_orders'] = related_orders
            
            if instruments is not None and trade.instrument in instruments:
                open_trades[trade.id] = Trade(**new_trade)
            elif instruments is None:
                open_trades[trade.id] = Trade(**new_trade)
        
        return open_trades
    
    
    def get_trade_details(self, trade_ID: int):
        """Returns the details of the trade specified by trade_ID.
        """
        
        response = self.api.trade.list(accountID=self.ACCOUNT_ID, ids=int(trade_ID))
        trade = response.body['trades'][0]
        
        details = {'direction': int(np.sign(trade.currentUnits)), 
                   'order_time': datetime.datetime.strptime(trade.openTime[:-4],
                                                            '%Y-%m-%dT%H:%M:%S.%f'), 
                   'instrument': trade.instrument, 
                   'size': trade.currentUnits,
                   'order_price': trade.price, 
                   'order_ID': trade.id, 
                   'time_filled': trade.openTime, 
                   'entry_price': trade.price, 
                   'unrealised_PL': trade.unrealizedPL, 
                   'margin_required': trade.marginUsed}
        
        # Get associated trades
        related = []
        try:
            details['take_profit'] = trade.takeProfitOrder.price
            related.append(trade.takeProfitOrder.id)
        except:
            pass
        
        try:
            details['stop_loss'] = trade.stopLossOrder.price
            related.append(trade.stopLossOrder.id)
        except:
            pass
        details['related_orders'] = related
        
        # TODO - veryify functionality of below...
        return Trade(trade)
    
    
    def get_positions(self, instrument: str = None, **kwargs) -> dict:
        """Gets the current positions open on the account. 
        """
        self._check_connection()
        response = self.api.position.list_open(accountID=self.ACCOUNT_ID)
        oanda_open_positions = response.body['positions']
        open_positions = {}
        for position in oanda_open_positions:
            pos = {'instrument': position.instrument,
                   'long_units': position.long.units,
                   'long_PL': position.long.unrealizedPL,
                   'long_margin': None,
                   'short_units': position.short.units,
                   'short_PL': position.short.unrealizedPL,
                   'short_margin': None,
                   'total_margin': position.marginUsed}
            
            # fetch trade ID'strade_IDs
            trade_IDs = []
            if abs(pos['long_units']) > 0: 
                for ID in position.long.tradeIDs: trade_IDs.append(ID)
            if abs(pos['short_units']) > 0: 
                for ID in position.short.tradeIDs: trade_IDs.append(ID)
            
            pos['trade_IDs'] = trade_IDs
            
            if instrument is not None and position.instrument == instrument:
                open_positions[position.instrument] = Position(**pos)
            elif instrument is None:
                open_positions[position.instrument] = Position(**pos)
        
        return open_positions
    
    
    def get_position(self, instrument: str) -> Position:
        """Gets position from Oanda.
        """
        self._check_connection()
        response = self.api.position.get(instrument = instrument, 
                                         accountID = self.ACCOUNT_ID)
        # TODO - convert to position
        return response.body['position']
    
    
    def get_summary(self):
        """Returns account summary.
        """
        self._check_connection()
        # response = self.api.account.get(accountID=self.ACCOUNT_ID)
        response = self.api.account.summary(accountID=self.ACCOUNT_ID)
        return response
    
    
    def get_data(self, pair: str, period: int, interval: str) -> pd.DataFrame:
        self._check_connection()
        response = self.api.instrument.candles(pair, granularity=interval,
                                               count=period, dailyAlignment = 0)
        data = self.utils.response_to_df(response)
        return data
    
    
    def check_trade_size(self, instrument: str, units: float) -> float:
        """Checks the requested trade size against the minimum trade size 
        allowed for the currency pair.
        """
        response = self.api.account.instruments(accountID=self.ACCOUNT_ID, 
                                                instruments = instrument)
        # minimum_units = response.body['instruments'][0].minimumTradeSize
        trade_unit_precision = response.body['instruments'][0].tradeUnitsPrecision
        return round(units, trade_unit_precision)
    
    
    def update_data(self, instrument: str, granularity: str, 
                    data: pd.DataFrame) -> pd.DataFrame:
        """Attempts to construct the latest candle when there is a delay in the 
        api feed.
        """
        self._check_connection()
        granularity_details = self.deconstruct_granularity(granularity)
        secs = granularity_details['seconds']
        mins = granularity_details['minutes']
        hrs  = granularity_details['hours']
        days = granularity_details['days']
        
        small_granularity = self.get_reduced_granularity(granularity_details, 25)
        
        # Get data equivalent of last candle's granularity
        time_now = datetime.datetime.now()
        start_time = time_now - datetime.timedelta(seconds = secs,
                                                   minutes = mins,
                                                   hours = hrs,
                                                   days = days)
        latest_data = self.get_historical_data(instrument, 
                                               small_granularity, 
                                               start_time.timestamp(), 
                                               time_now.timestamp())
        
        # Get latest price data
        latest_close = latest_data.Close.values[0]
        
        open_price = data.Close.values[-1]
        close_price = latest_close
        high_price = max(latest_data.High.values)
        low_price = min(latest_data.Low.values)
        last_time = data.index[-1]
        stripped_time = datetime.datetime.strptime(last_time.strftime("%Y-%m-%d %H:%M:%S%z"),
                                                      "%Y-%m-%d %H:%M:%S%z")
        new_time = stripped_time + datetime.timedelta(seconds = secs,
                                                      minutes = mins,
                                                      hours = hrs,
                                                      days = days)
        
        new_candle = pd.DataFrame({'Open'  : open_price, 
                                   'High'  : high_price,
                                   'Low'   : low_price,
                                   'Close' : close_price},
                                  index=[new_time])
        
        new_data = pd.concat([data, new_candle])
        
        return new_data
    
    
    def get_historical_data(self, instrument, interval, from_time, to_time):
        
        self._check_connection()
        
        response        = self.api.instrument.candles(instrument,
                                                      granularity = interval,
                                                      fromTime = from_time,
                                                      toTime = to_time
                                                      )
        
        data = self.utils.response_to_df(response)
        
        return data
    
    
    def deconstruct_granularity(self, granularity: str):
        """Returns a dict with the seconds, minutes, hours and days
        corresponding to a granularity. 
        """
        
        # Get letter to determine timeframe (eg. M)
        letter = granularity[0]
        
        # Get timeframe multiple (eg. 15)
        if len(granularity) > 1:
            number = float(granularity[1:])
        else:
            number = 1
        
        
        if letter == 'S':
            seconds     = number
            minutes     = 0
            hours       = 0
            days        = 0
        
        elif letter == 'M':
            seconds     = 0
            minutes     = number
            hours       = 0
            days        = 0
            
        elif letter == 'H':
            seconds     = 0
            minutes     = 0
            hours       = number
            days        = 0
            
        else:
            seconds     = 0
            minutes     = 0
            hours       = 0
            days        = number
        
        granularity_details = {'seconds': seconds,
                               'minutes': minutes,
                               'hours': hours,
                               'days': days}
        
        return granularity_details


    def get_reduced_granularity(self, granularity_details, fraction):
        '''Returns a candlestick granularity as a fraction of given granularity'''
        secs = granularity_details['seconds']
        mins = granularity_details['minutes']
        hrs  = granularity_details['hours']
        days = granularity_details['days']
        
        total_seconds = secs + 60*mins + 60*60*hrs + 60*60*24*days
        
        fractional_seconds = total_seconds/fraction
        
        seconds = fractional_seconds
        minutes = fractional_seconds/60
        hours = fractional_seconds/(60*60)
        days = fractional_seconds/(60*60*24)
        
        if days > 1:
            letter = 'D'
            number = 1
            reduced_granularity = letter
            
        elif hours > 1:
            base   = 2
            letter = 'H'
            number = base*round(hours/base)
            if number > 12:
                number = 12
            reduced_granularity = letter + str(number)
                
        elif minutes > 1:
            base   = 15 
            letter = 'M'
            number = base*round(minutes/base)
            if number > 30:
                number = 30
            reduced_granularity = letter + str(number)
            
        else: 
            base   = 15 
            letter = 'S'
            number = base*round(seconds/base)
            if number > 30:
                number = 30
            reduced_granularity = letter + str(number)
        
        if reduced_granularity[1:] == '0':
            reduced_granularity = reduced_granularity[0] + '1'
        
        return reduced_granularity
    
    
    def get_pip_location(self, instrument: str):
        """Returns the pip location of the requested instrument.
        """
        response = self.api.account.instruments(self.ACCOUNT_ID, 
                                                instruments=instrument)
        return response.body['instruments'][0].pipLocation
    
    
    def get_trade_unit_precision(self, instrument: str):
        """Returns the trade unit precision for the requested instrument.
        """
        # TODO - implement checking using this method
        response = self.api.account.instruments(self.ACCOUNT_ID, 
                                                instruments=instrument)
        return response.body['instruments'][0].tradeUnitsPrecision
    
    
    def _check_connection(self) -> None:
        """Connects to Oanda v20 REST API. An initial call is performed to check
        for a timeout error.
        """
        # TODO - improve this - currently doubles the poll rate
        for atempt in range(10):
            try:
                # Attempt basic task to check connection
                self.api.account.get(accountID=self.ACCOUNT_ID)
            
            except BaseException:
                # Error has occurred
                ex_type, ex_value, ex_traceback = sys.exc_info()
            
                # Extract unformatter stack traces as tuples
                trace_back = traceback.extract_tb(ex_traceback)
            
                # Format stacktrace
                stack_trace = list()
            
                for trace in trace_back:
                    trade_string = "File : %s , Line : %d, " % (trace[0], trace[1]) + \
                                   "Func.Name : %s, Message : %s" % (trace[2], trace[3])
                    stack_trace.append(trade_string)
                
                print("\nWARNING FROM OANDA API: The following exception was caught.")
                print("Time: {}".format(datetime.datetime.now().strftime("%b %d %H:%M:%S")))
                print("Exception type : %s " % ex_type.__name__)
                print("Exception message : %s" %ex_value)
                print("Stack trace : %s" %stack_trace)
                print("  Attempting to reconnect to Oanda v20 API.")
                
                time.sleep(3)
                api = v20.Context(hostname = self.API, 
                                  token = self.ACCESS_TOKEN, 
                                  port = self.port)
                self.api = api
            
            else:
                break
            
        else:
            print("FATAL: All attempts to connect to Oanda API have failed.")
        
        
    def _place_market_order(self, order: Order):
        """Places market order.
        """
        self._check_connection()
        stop_loss_details = self._get_stop_loss_details(order)
        take_profit_details = self._get_take_profit_details(order)
        
        # Check position size
        size = self.check_trade_size(order.instrument, 
                                     order.size)
        
        response = self.api.order.market(accountID = self.ACCOUNT_ID,
                                         instrument = order.instrument,
                                         units = order.direction * size,
                                         takeProfitOnFill = take_profit_details,
                                         stopLossOnFill = stop_loss_details,)
        return response
    
    
    def _place_stop_limit_order(self, order):
        """Places MarketIfTouchedOrder with Oanda.
        https://developer.oanda.com/rest-live-v20/order-df/
        """
        self._check_connection()
        
        stop_loss_details = self._get_stop_loss_details(order)
        take_profit_details = self._get_take_profit_details(order)
        
        # Check and correct order stop price
        price = self._check_precision(order.instrument, 
                                      order.order_stop_price)
        
        trigger_condition = order.trigger_price
        
        # Need to test cases when no stop/take is provided (as None type)
        # TODO - support other order types (see link above, market is default)
        response = self.api.order.market_if_touched(accountID = self.ACCOUNT_ID,
                                                    instrument = order.instrument,
                                                    units = order.direction * order.size,
                                                    price = str(price),
                                                    takeProfitOnFill = take_profit_details,
                                                    stopLossOnFill = stop_loss_details,
                                                    triggerCondition = trigger_condition)
        return response
    
    
    def _place_limit_order(self, order: Order):
        """(NOT YET IMPLEMENTED) PLaces limit order. 
        """
        raise Exception("Limit orders are not yet implemented for Oanda. "+\
                        "Please raise an issue on GitHub.")
        

    def _get_stop_loss_details(self, order: Order) -> dict:
        """Constructs stop loss details dictionary.
        """
        self._check_connection()
        if order.stop_type is not None:
            price = self._check_precision(order.instrument, order.stop_loss)
            
            if order.stop_type == 'trailing':
                # Trailing stop loss order
                stop_loss_details = {"price": str(price),
                                     "type": "TRAILING_STOP_LOSS"}
            else:
                stop_loss_details = {"price": str(price)}
        else:
            stop_loss_details = None
        
        return stop_loss_details
    
    
    def _get_take_profit_details(self, order: Order) -> dict:
        """Constructs take profit details dictionary.
        """
        self._check_connection()
        if order.take_profit is not None:
            price = self._check_precision(order.instrument, order.take_profit)
            take_profit_details = {"price": str(price)}
        else:
            take_profit_details = None
        
        return take_profit_details


    def _check_response(self, response):
        """Checks API response (currently only for placing orders).
        """
        if response.status != 201:
            message = response.body['errorMessage']
        else:
            message = "Success."
            
        output = {'Status': response.status, 
                  'Message': message}
        # TODO - print errors
        return output
    
    
    def _close_position(self, instrument, long_units=None, short_units=None,
                       **kwargs):
        """Closes all open positions on an instrument.
        """
        self._check_connection()
        # Check if the position is long or short
        # Temp code to close all positions
        # Close all long units
        response = self.api.position.close(accountID=self.ACCOUNT_ID, 
                                           instrument=instrument,
                                           longUnits="ALL")
        
        # Close all short units
        response = self.api.position.close(accountID=self.ACCOUNT_ID, 
                                           instrument=instrument,
                                           shortUnits="ALL")
        
        # TODO - the code below makes no sense currently; specifically, 
        # position.long.Units ????

        # open_position = self.get_open_positions(instrument)        

        # if len(open_position) > 0:
        #     position = open_position['position']
            
        #     if long_units is None:
        #         long_units  = position.long.units
        #     if short_units is None:
        #         short_units = position.short.units
            
        #     if long_units > 0:
        #         response = self.api.position.close(accountID=self.ACCOUNT_ID, 
        #                                            instrument=instrument, 
        #                                            longUnits="ALL")
            
        #     elif short_units > 0: 
        #         response = self.api.position.close(accountID=self.ACCOUNT_ID, 
        #                                            instrument=instrument,
        #                                            shortUnits="ALL")
            
        #     else:
        #         print("There is no current position with {} to close.".format(instrument))
        #         response = None
        # else:
        #     response = None
            
        return response
    
    
    def _get_precision(self, instrument: str):
        """Returns the allowable precision for a given pair.
        """
        self._check_connection()
        response = self.api.account.instruments(accountID = self.ACCOUNT_ID, 
                                                instruments = instrument)
        precision = response.body['instruments'][0].displayPrecision
        return precision

    
    def _check_precision(self, instrument, price):
        """Modify a price based on required ordering precision for pair.
        """
        N = self._get_precision(instrument)
        corrected_price = round(price, N)
        return corrected_price
    
    
    def _get_order_book(self, instrument: str):
        """Returns the order book of the instrument specified. 
        """
        response = self.api.instrument.order_book(instrument)
        return response.body['orderBook']
        
    
    def _get_position_book(self, instrument: str):
        """Returns the position book of the instrument specified. 
        """
        response = self.api.instrument.position_book(instrument)
        return response.body['positionBook']
    
    
    


import os
import sys
from getopt import getopt
from autotrader.utilities import read_yaml
from autotrader.emailing import send_order_summary


def main(uo_dict: dict) -> None:
    email_dir = os.path.dirname(os.path.abspath(__file__))
    file_dir = os.path.join(email_dir, '../logfiles')
    file_path = os.path.join(file_dir, uo_dict["filename"])
    
    global_config = read_yaml(email_dir + '/../config/GLOBAL.yaml')
    host_email = global_config["EMAILING"]["HOST_ACCOUNT"]
    mailing_list = global_config["EMAILING"]["MAILING_LIST"]
    
    send_order_summary(file_path, mailing_list, host_email)


short_options = "f:"
long_options = ['file=']


if __name__ == '__main__':
    options, r = getopt(sys.argv[1:], 
                          short_options, 
                          long_options
                          )
    
    # Defaults
    filename = None
    
    for opt, arg in options:
        if opt in ('-f', '--file'):
            filename = arg
        
    uo_dict = {'filename': filename}

    main(uo_dict)



import numpy as np
import pandas as pd
from finta import TA
from typing import Union


def supertrend(data: pd.DataFrame, period: int = 10, ATR_multiplier: float = 3.0, 
               source: pd.Series = None) -> pd.DataFrame:
    """SuperTrend indicator, ported from the SuperTrend indicator by 
    KivancOzbilgic on TradingView.

    Parameters
    ----------
    data : pd.DataFrame
        The OHLC data.
    period : int, optional
        The lookback period. The default is 10.
    ATR_multiplier : int, optional
        The ATR multiplier. The default is 3.0.
    source : pd.Series, optional
        The source series to use in calculations. If None, hl/2 will be 
        used. The default is None.

    Returns
    -------
    supertrend_df : pd.DataFrame
        A Pandas DataFrame of containing the SuperTrend indicator, with 
        columns of 'uptrend' and 'downtrend' containing uptrend/downtrend
        support/resistance levels, and 'trend', containing -1/1 to indicate
        the current implied trend. 

    References
    ----------
    https://www.tradingview.com/script/r6dAP7yi/
    """
    
    if source is None:
        source = (data.High.values + data.Low.values) / 2
    
    # Calculate ATR
    atr = TA.ATR(data, period)
    
    up = source - (ATR_multiplier*atr)
    up_list = [up[0]]
    up_times = [data.index[0]]
    N_up = 0
    
    dn = source + (ATR_multiplier*atr)
    dn_list = [dn[0]]
    dn_times = [data.index[0]]
    N_dn = 0
    
    trend = 1
    trend_list = [trend]
    
    for i in range(1, len(data)):
        if trend == 1:
            if data.Close.values[i] > max(up[N_up:i]):
                up_list.append(max(up[N_up:i]))
                up_times.append(data.index[i])
                
                dn_list.append(np.nan)
                dn_times.append(data.index[i])
            else: 
                trend = -1
                N_dn = i
                dn_list.append(dn[i])
                dn_times.append(data.index[i])
                
                up_list.append(np.nan)
                up_times.append(data.index[i])
                
        else:
            if data.Close.values[i] < min(dn[N_dn:i]):
                dn_list.append(min(dn[N_dn:i]))
                dn_times.append(data.index[i])
                
                up_list.append(np.nan)
                up_times.append(data.index[i])
            else:
                trend = 1
                N_up = i
                up_list.append(up[i])
                up_times.append(data.index[i])
                
                dn_list.append(np.nan)
                dn_times.append(data.index[i])
        
        trend_list.append(trend)
    
    supertrend_df = pd.DataFrame({'uptrend': up_list,
                                  'downtrend': dn_list,
                                  'trend': trend_list}, 
                                 index = up_times)
    return supertrend_df


def halftrend(data: pd.DataFrame, amplitude: int = 2, 
               channel_deviation: float = 2) -> pd.DataFrame:
    """HalfTrend indicator, ported from the HalfTrend indicator by 
    Alex Orekhov (everget) on TradingView.

    Parameters
    ----------
    data : pd.DataFrame
        OHLC price data.
    amplitude : int, optional
        The lookback window. The default is 2.
    channel_deviation : float, optional
        The ATR channel deviation factor. The default is 2.

    Returns
    -------
    htdf : TYPE
        DESCRIPTION.

    References
    ----------
    https://www.tradingview.com/script/U1SJ8ubc-HalfTrend/
    """
    
    # Initialisation
    atr2 = TA.ATR(data, 100)/2
    dev = channel_deviation * atr2
    high_price = data.High.rolling(amplitude).max().fillna(0)
    low_price = data.Low.rolling(amplitude).min().fillna(0)
    highma = TA.SMA(data, period=amplitude, column='High')
    lowma = TA.SMA(data, period=amplitude, column='Low')
    
    trend = np.zeros(len(data))
    next_trend = np.zeros(len(data))
    max_low_price = np.zeros(len(data))
    max_low_price[0] = data.Low[0]
    min_high_price = np.zeros(len(data))
    min_high_price[0] = data.High[0]
    
    for i in range(1, len(data)):
        if next_trend[i-1] == 1:
            max_low_price[i] = max(low_price[i-1], max_low_price[i-1])

            if highma[i] < max_low_price[i] and data.Close[i] < data.Low[i-1]:
                trend[i] = 1
                next_trend[i] = 0
                min_high_price[i] = high_price[i]
            else:
                # assign previous values again
                trend[i] = trend[i-1]
                next_trend[i] = next_trend[i-1]
                min_high_price[i] = min_high_price[i-1]
        else:
            min_high_price[i] = min(high_price[i-1], min_high_price[i-1])
            
            if lowma[i] > min_high_price[i] and data.Close[i] > data.High[i-1]:
                trend[i] = 0
                next_trend[i] = 1
                max_low_price[i] = low_price[i]
            else:
                # assign previous values again
                trend[i] = trend[i-1]
                next_trend[i] = next_trend[i-1]
                max_low_price[i] = max_low_price[i-1]
                
    up = np.zeros(len(data))
    up[0] = max_low_price[0]
    down = np.zeros(len(data))
    down[0] = min_high_price[0]
    atr_high = np.zeros(len(data))
    atr_low = np.zeros(len(data))
    
    for i in range(1, len(data)):
        if trend[i] == 0:
            if trend[i-1] != 0:
                up[i] = down[i-1]
            else:
                up[i] = max(max_low_price[i-1], up[i-1])
            
            atr_high[i] = up[i] + dev[i]
            atr_low[i] = up[i] - dev[i]
            
        else:
            if trend[i-1] != 1:
                down[i] = up[i-1]
            else:
                down[i] = min(min_high_price[i-1], down[i-1]) 
                
            atr_high[i] = down[i] + dev[i]
            atr_low[i] = down[i] - dev[i]
    
    halftrend = np.where(trend == 0, up, down)
    buy = np.where((trend==0) & (np.roll(trend,1)==1), 1, 0)
    sell = np.where((trend==1) & (np.roll(trend,1)==0), 1, 0)
    
    # Construct DataFrame
    htdf = pd.DataFrame(data = {'halftrend': halftrend, 
                                'atrHigh': np.nan_to_num(atr_high),
                                'atrLow': np.nan_to_num(atr_low),
                                'buy': buy, 
                                'sell': sell},
                        index = data.index)
    
    # Clear false leading signals
    htdf.buy.values[:100] = np.zeros(100)
    htdf.sell.values[:100] = np.zeros(100)
    
    # Replace leading zeroes with nan
    htdf['atrHigh'] = htdf.atrHigh.replace(to_replace=0, value=float("nan"))
    htdf['atrLow'] = htdf.atrLow.replace(to_replace=0, value=float("nan"))
    
    return htdf


def range_filter(data: pd.DataFrame, range_qty: float = 2.618, 
                 range_period: int = 14, smooth_range: bool = True, 
                 smooth_period: int = 27, av_vals: bool = False, 
                 av_samples: int = 2, mov_source: str = 'body', 
                 filter_type: int = 1) -> pd.DataFrame:
    """Price range filter, ported from the Range Filter [DW] indicator by
    DonovanWall on TradingView. The indicator was designed to filter out 
    minor price action for a clearer view of trends.

    Parameters
    ----------
    data : pd.DataFrame
        The OHLC price data.
    range_qty : float, optional
        The range size. The default is 2.618.
    range_period : int, optional
        The range period. The default is 14.
    smooth_range : bool, optional
        Smooth the price range. The default is True.
    smooth_period : int, optional
        The smooting period. The default is 27.
    av_vals : bool, optional
        Average values. The default is False.
    av_samples : int, optional
        The number of average samples to use. The default is 2.
    mov_source : str, optional
        The price movement source ('body' or 'wicks'). The default is 'body'.
    filter_type : int, optional
        The filter type to use in calculations (1 or 2). The default is 1.

    Returns
    -------
    rfi : pd.DataFrame
        A dataframe containing the range filter indicator bounds.
        
    References
    ----------
    https://www.tradingview.com/script/lut7sBgG-Range-Filter-DW/
    """
    # Get high and low values
    if mov_source == 'body':
        high_val = data.Close
        low_val = data.Close
    elif mov_source == 'wicks':
        high_val = data.High
        low_val = data.Low
    
    # Get filter values
    rng = _range_size((high_val + low_val)/2, 'AverageChange', 
                      range_qty, range_period)
    rfi = _calculate_range_filter(high_val, low_val, rng, range_period, filter_type, 
                       smooth_range, smooth_period, av_vals, av_samples)
    
    return rfi


def bullish_engulfing(data: pd.DataFrame, detection: str = None):
    """Bullish engulfing pattern detection. 
    """
    if detection == "SMA50":
        sma50       = sma(data.Close.values, 50)
        down_trend  = np.where(data.Close.values < sma50, True, False)
        
    elif detection == "SMA50/200":
        sma50       = sma(data.Close.values, 50)
        sma200      = sma(data.Close.values, 200)
        
        down_trend  = np.where((data.Close.values < sma50) & 
                               (data.Close.values < sma200), 
                               True, False)
    else:
        down_trend  = np.full(len(data), True)
    
    body_len        = 14    # ema depth for bodyAvg
    
    body_high       = np.maximum(data.Close.values, data.Open.values)
    body_low        = np.minimum(data.Close.values, data.Open.values)
    body            = body_high - body_low
    
    body_avg        = ema(body, body_len)
    short_body      = body < body_avg
    long_body       = body > body_avg
    
    white_body      = data.Open.values < data.Close.values
    black_body      = data.Open.values > data.Close.values
    
    inside_bar = [False]
    for i in range(1, len(data)):
        val  = (body_high[i-1] > body_high[i]) and (body_low[i-1] < body_low[i])
        inside_bar.append(val)
        
    engulfing_bullish = [False]
    for i in range(1, len(data)):
        condition = down_trend[i] & \
                    white_body[i] & \
                    long_body[i] & \
                    black_body[i-1] & \
                    short_body[i-1] & \
                    (data.Close.values[i] >= data.Open.values[i-1]) & \
                    (data.Open.values[i] <= data.Close.values[i-1]) & \
                    ((data.Close.values[i] > data.Open.values[i-1]) | (data.Open.values[i] < data.Close.values[i-1]))
        
        engulfing_bullish.append(condition)
        
    return engulfing_bullish


def bearish_engulfing(data: pd.DataFrame, detection: str = None):
    """Bearish engulfing pattern detection. 
    """    
    if detection == "SMA50":
        sma50       = sma(data.Close.values, 50)
        up_trend    = np.where(data.Close.values > sma50, True, False)
    elif detection == "SMA50/200":
        sma50       = sma(data.Close.values, 50)
        sma200      = sma(data.Close.values, 200)
        
        up_trend    = np.where((data.Close.values > sma50) & 
                               (data.Close.values > sma200), 
                               True, False)
    else:
        up_trend    = np.full(len(data), True)
    
    body_len        = 14    # ema depth for bodyAvg
    body_high       = np.maximum(data.Close.values, data.Open.values)
    body_low        = np.minimum(data.Close.values, data.Open.values)
    body            = body_high - body_low
    
    body_avg        = ema(body, body_len)
    short_body      = body < body_avg
    long_body       = body > body_avg
    
    white_body      = data.Open.values < data.Close.values
    black_body      = data.Open.values > data.Close.values
    
    inside_bar = [False]
    for i in range(1, len(data)):
        val  = (body_high[i-1] > body_high[i]) and (body_low[i-1] < body_low[i])
        inside_bar.append(val)
        
    engulfing_bearish = [False]
    for i in range(1, len(data)):
        condition = up_trend[i] & \
                    black_body[i] & \
                    long_body[i] & \
                    white_body[i-1] & \
                    short_body[i-1] & \
                    (data.Close.values[i] <= data.Open.values[i-1]) & \
                    (data.Open.values[i] >= data.Close.values[i-1]) & \
                    ((data.Close.values[i] < data.Open.values[i-1]) | (data.Open.values[i] > data.Close.values[i-1]))
        
        engulfing_bearish.append(condition)
        
    return engulfing_bearish


def find_swings(data: pd.DataFrame, data_type: str = 'ohlc', 
                n: int = 2) -> pd.DataFrame:
    """Locates swings in the inputted data using a moving average gradient
    method.

    Parameters
    ----------
    data : pd.DataFrame
        An OHLC dataframe of price, or an array/list/Series of data from an 
        indicator (eg. RSI).
    data_type : str, optional
        The nature of the data ('ohlc' or 'other'). The default is 'ohlc'.
    n : int, optional
        The moving average period. The default is 2.

    Returns
    -------
    swing_df : pd.DataFrame
        A dataframe containing the swing levels detected.
    """
    # Prepare data 
    if data_type == 'ohlc':
        # Find swings in OHLC price data
        # hl2     = (data.Open.values + data.Close.values)/2
        hl2     = (data.High.values + data.Low.values)/2
        swing_data = ema(hl2, n)
        
        low_data = data.Low.values
        high_data = data.High.values
        
    else:
        # Find swings in alternative data source
        swing_data = data
        low_data = data
        high_data = data
    
    # Calculate slope of data and points where slope changes
    grad = [swing_data[i] - swing_data[i-1] for i in range(len(swing_data))]
    swings = np.where(np.sign(grad) != np.sign(np.roll(grad,1)), -np.sign(grad), 0)

    # Construct columns
    low_list    = [0,0]
    high_list   = [0,0]
    for i in range(2,len(data)):
        if swings[i] == -1:
            # Down swing - find min price in the vicinity 
            high_list.append(0)
            low_list.append(min(low_data[i-n:i]))
            
        elif swings[i] == 1:
            # Up swing - find max price in the vicinity
            high_list.append(max(high_data[i-n:i]))
            low_list.append(0)
            
        else:
            # Price movement
            low_list.append(0)
            high_list.append(0)
    
    trend = rolling_signal_list(-swings)
    swings_list = merge_signals(low_list, high_list)
    last_swing = rolling_signal_list(swings_list)
    last_swing[0:n] = list(high_data[0:n])
    
    # Need to return both a last swing low and last swing high list
    last_low = rolling_signal_list(low_list)
    last_low[0:n] = list(low_data[0:n])     # Fill start of data
    last_high = rolling_signal_list(high_list)
    last_high[0:n] = list(high_data[0:n])   # Fill start of data
    
    swing_df = pd.DataFrame(data={'Highs': last_high, 
                                  'Lows' : last_low,
                                  'Last' : last_swing,
                                  'Trend': trend},
                            index = data.index)
    
    return swing_df
    

def classify_swings(swing_df: pd.DataFrame, tol: int = 0) -> pd.DataFrame:
    """Classifies a dataframe of swings (from find_swings) into higher-highs, 
    lower-highs, higher-lows and lower-lows.
    

    Parameters
    ----------
    swing_df : pd.DataFrame
        The dataframe returned by find_swings.
    tol : int, optional
        The classification tolerance. The default is 0.

    Returns
    -------
    swing_df : pd.DataFrame
        A dataframe containing the classified swings.
    """
    # Create copy of swing dataframe
    swing_df = swing_df.copy()
    
    new_level = np.where(swing_df.Last != swing_df.Last.shift(), 1, 0)

    candles_since_last = candles_between_crosses(new_level)
    
    # Add column 'candles since last swing' CSLS
    swing_df['CSLS'] = candles_since_last
    
    # Find strong Support and Resistance zones
    swing_df['Support'] = (swing_df.CSLS > tol) & (swing_df.Trend == 1)
    swing_df['Resistance'] = (swing_df.CSLS > tol) & (swing_df.Trend == -1)
    
    # Find higher highs and lower lows
    swing_df['Strong_lows'] = swing_df['Support'] * swing_df['Lows'] # Returns high values when there is a strong support
    swing_df['Strong_highs'] = swing_df['Resistance'] * swing_df['Highs'] # Returns high values when there is a strong support
    
    # Remove duplicates to preserve indexes of new levels
    swing_df['FSL'] = unroll_signal_list(swing_df['Strong_lows']) # First of new strong lows
    swing_df['FSH'] = unroll_signal_list(swing_df['Strong_highs']) # First of new strong highs
    
    # Now compare each non-zero value to the previous non-zero value.
    low_change = np.sign(swing_df.FSL) * (swing_df.FSL - swing_df.Strong_lows.replace(to_replace=0, method='ffill').shift())
    high_change = np.sign(swing_df.FSH) * (swing_df.FSH - swing_df.Strong_highs.replace(to_replace=0, method='ffill').shift())
    
    swing_df['LL'] = np.where(low_change < 0, True, False)
    swing_df['HL'] = np.where(low_change > 0, True, False)
    swing_df['HH'] = np.where(high_change > 0, True, False)
    swing_df['LH'] = np.where(high_change < 0, True, False)
    
    return swing_df


def detect_divergence(classified_price_swings: pd.DataFrame, 
                      classified_indicator_swings: pd.DataFrame, tol: int = 2,
                      method: int = 0) -> pd.DataFrame:
    """Detects divergence between price swings and swings in an indicator.

    Parameters
    ----------
    classified_price_swings : pd.DataFrame
        The output from classify_swings using OHLC data.
    classified_indicator_swings : pd.DataFrame
        The output from classify_swings using indicator data.
    tol : int, optional
        The number of candles which conditions must be met within. The 
        default is 2.
    method : int, optional
        The method to use when detecting divergence (0 or 1). The default is 0.

    Raises
    ------
    Exception
        When an unrecognised method of divergence detection is requested.

    Returns
    -------
    divergence : pd.DataFrame
        A dataframe containing divergence signals.
        
    Notes
    -----
    Options for the method include:
        0: use both price and indicator swings to detect divergence (default)
        
        1: use only indicator swings to detect divergence (more responsive)
    """
    if method == 0:
        regular_bullish = []
        regular_bearish = []
        hidden_bullish = []
        hidden_bearish = []
        
        for i in range(len(classified_price_swings)):
            # Look backwards in each
            
            # REGULAR BULLISH DIVERGENCE
            if sum(classified_price_swings['LL'][i-tol:i]) + sum(classified_indicator_swings['HL'][i-tol:i]) > 1:
                regular_bullish.append(True)
            else:
                regular_bullish.append(False)
            
            # REGULAR BEARISH DIVERGENCE
            if sum(classified_price_swings['HH'][i-tol:i]) + sum(classified_indicator_swings['LH'][i-tol:i]) > 1:
                regular_bearish.append(True)
            else:
                regular_bearish.append(False)
            
            # HIDDEN BULLISH DIVERGENCE
            if sum(classified_price_swings['HL'][i-tol:i]) + sum(classified_indicator_swings['LL'][i-tol:i]) > 1:
                hidden_bullish.append(True)
            else:
                hidden_bullish.append(False)
            
            # HIDDEN BEARISH DIVERGENCE
            if sum(classified_price_swings['LH'][i-tol:i]) + sum(classified_indicator_swings['HH'][i-tol:i]) > 1:
                hidden_bearish.append(True)
            else:
                hidden_bearish.append(False)
            
            divergence = pd.DataFrame(data = {'regularBull': unroll_signal_list(regular_bullish),
                                              'regularBear': unroll_signal_list(regular_bearish),
                                              'hiddenBull': unroll_signal_list(hidden_bullish),
                                              'hiddenBear': unroll_signal_list(hidden_bearish)})
    elif method == 1:
        # Use indicator swings only to detect divergence
        for i in range(len(classified_price_swings)):
            
            price_at_indi_lows = (classified_indicator_swings['FSL'] != 0) * classified_price_swings['Lows']
            price_at_indi_highs = (classified_indicator_swings['FSH'] != 0) * classified_price_swings['Highs']
            
            # Determine change in price between indicator lows
            price_at_indi_lows_change = np.sign(price_at_indi_lows) * (price_at_indi_lows - price_at_indi_lows.replace(to_replace=0, method='ffill').shift())
            price_at_indi_highs_change = np.sign(price_at_indi_highs) * (price_at_indi_highs - price_at_indi_highs.replace(to_replace=0, method='ffill').shift())
            
            
            # DETECT DIVERGENCES
            regular_bullish = (classified_indicator_swings['HL']) & (price_at_indi_lows_change < 0 )
            regular_bearish = (classified_indicator_swings['LH']) & (price_at_indi_highs_change > 0 )
            hidden_bullish = (classified_indicator_swings['LL']) & (price_at_indi_lows_change > 0 )
            hidden_bearish = (classified_indicator_swings['HH']) & (price_at_indi_highs_change < 0 )
            
            divergence = pd.DataFrame(data = {'regularBull': regular_bullish,
                                              'regularBear': regular_bearish,
                                              'hiddenBull': hidden_bullish,
                                              'hiddenBear': hidden_bearish})
        
    else:
        raise Exception("Error: unrecognised method of divergence detection.")
        
    return divergence


def autodetect_divergence(ohlc: pd.DataFrame, indicator_data: pd.DataFrame, 
                          tolerance: int = 1, method: int = 0) -> pd.DataFrame:
    """A wrapper method to automatically detect divergence from inputted OHLC price 
    data and indicator data.

    Parameters
    ----------
    ohlc : pd.DataFrame
        A dataframe of OHLC price data.
    indicator_data : pd.DataFrame
        dataframe of indicator data.
    tolerance : int, optional
        A parameter to control the lookback when detecting divergence. 
        The default is 1.
    method : int, optional
        The divergence detection method. Set to 0 to use both price and 
        indicator swings to detect divergence. Set to 1 to use only indicator 
        swings to detect divergence. The default is 0.

    Returns
    -------
    divergence : pd.DataFrame
        A DataFrame containing columns 'regularBull', 'regularBear',
        'hiddenBull' and 'hiddenBear'.

    See Also
    --------
    autotrader.indicators.find_swings
    autotrader.indicators.classify_swings
    autotrader.indicators.detect_divergence
    
    """
    
    # Price swings
    price_swings = find_swings(ohlc)
    price_swings_classified = classify_swings(price_swings)
    
    # Indicator swings 
    indicator_swings = find_swings(indicator_data, data_type='other')
    indicator_classified = classify_swings(indicator_swings)
    
    # Detect divergence
    divergence = detect_divergence(price_swings_classified, indicator_classified, 
                                   tol=tolerance, method=method)
    
    return divergence


def heikin_ashi(data: pd.DataFrame):
    """Calculates the Heikin-Ashi candlesticks from Japanese candlestick 
    data. 
    """
    # Create copy of data to prevent overwriting
    working_data = data.copy()
    
    # Calculate Heikin Ashi candlesticks
    ha_close = 0.25*(working_data.Open + working_data.Low + working_data.High + working_data.Close)
    ha_open = 0.5*(working_data.Open + working_data.Close)
    ha_high = np.maximum(working_data.High.values, working_data.Close.values, working_data.Open.values)
    ha_low = np.minimum(working_data.Low.values, working_data.Close.values, working_data.Open.values)
    
    ha_data = pd.DataFrame(data={'Open': ha_open, 
                                 'High': ha_high, 
                                 'Low': ha_low, 
                                 'Close': ha_close}, 
                           index=working_data.index)
    
    return ha_data
 
    
def ha_candle_run(ha_data: pd.DataFrame):
    """Returns a list for the number of consecutive green and red 
    Heikin-Ashi candles.
    
    Parameters
    ----------
    ha_data: pd.DataFrame
        The Heikin Ashi OHLC data.
    
    See Also
    --------
    heikin_ashi
    """    
    green_candle = np.where(ha_data.Close - ha_data.Open > 0, 1, 0)
    red_candle = np.where(ha_data.Close - ha_data.Open < 0, 1, 0)
    
    green_run = []
    red_run = []
    
    green_sum = 0
    red_sum = 0
    
    for i in range(len(ha_data)):
        if green_candle[i] == 1:
            green_sum += 1
        else:
            green_sum = 0
        
        if red_candle[i] == 1:
            red_sum += 1
        else:
            red_sum = 0
        
        green_run.append(green_sum)
        red_run.append(red_sum)
        
    return green_run, red_run


def N_period_high(data: pd.DataFrame, N: int):
    """Returns the N-period high. 
    """
    highs = data.High.rolling(N).max()
    return highs


def N_period_low(data: pd.DataFrame, N: int):
    """Returns the N-period low.
    """
    lows = data.Low.rolling(N).min()
    return lows


def crossover(ts1: pd.Series, ts2: pd.Series) -> pd.Series:
    """Locates where two timeseries crossover each other, returning 1 when
    list_1 crosses above list_2, and -1 for when list_1 crosses below list_2.

    Parameters
    ----------
    ts1 : pd.Series
        The first timeseries.
    ts2 : pd.Series
        The second timeseries.

    Returns
    -------
    crossovers : pd.Series
        The crossover series.
    """
    
    signs = np.sign(ts1 - ts2)
    crossovers = pd.Series(data=signs*(signs != signs.shift(1)), name='crossovers')
    
    return crossovers


def cross_values(ts1: Union[list, pd.Series], ts2: Union[list, pd.Series], 
                 ts_crossover: Union[list, pd.Series] = None) -> Union[list, pd.Series]:
    """Returns the approximate value of the point where the two series cross.

    Parameters
    ----------
    ts1 : list | pd.Series
        The first timeseries..
    ts2 : list | pd.Series
        The second timeseries..
    ts_crossover : list | pd.Series, optional
        The crossovers between timeseries 1 and timeseries 2.

    Returns
    -------
    cross_points : list | pd.Series
        The values at which crossovers occur.
    """
    
    if ts_crossover is None:
        ts_crossover = crossover(ts1, ts2)
    
    last_cross_point = ts1[0]
    cross_points = [last_cross_point]
    for i in range(1, len(ts_crossover)):
        if ts_crossover[i] != 0:
            i0 = 0
            m_a = ts1[i] - ts1[i-1]
            m_b = ts2[i] - ts2[i-1]
            ix = (ts2[i-1] - ts1[i-1])/(m_a-m_b) + i0
            
            cross_point = m_a*(ix - i0) + ts1[i-1]
            
            last_cross_point = cross_point
            
        else:
            cross_point = last_cross_point
        
        cross_points.append(cross_point)
    
    # Replace nans with 0
    cross_points = [0 if x!=x  else x for x in cross_points]
    
    if isinstance(ts1, pd.Series):
        # Convert to Series
        cross_points = pd.Series(data=cross_points, 
                                 index=ts1.index, 
                                 name='crossval')
    
    return cross_points


def candles_between_crosses(crosses: Union[list, pd.Series]) -> Union[list, pd.Series]:
    """Returns a rolling sum of candles since the last cross/signal occurred.

    Parameters
    ----------
    crosses : list | pd.Series
        The list or Series containing crossover signals.

    Returns
    -------
    counts : TYPE
        The rolling count of bars since the last crossover signal.
        
    See Also
    ---------
    indicators.crossover
    """
    
    count = 0
    counts = []
    
    for i in range(len(crosses)):
        if crosses[i] == 0:
            # Change in signal - reset count
            count += 1
        else:
            count = 0
        
        counts.append(count)
    
    if isinstance(crosses, pd.Series):
        # Convert to Series
        counts = pd.Series(data=counts, index=crosses.index, name='counts')
    
    return counts


def rolling_signal_list(signals: Union[list, pd.Series]) -> list:
    """Returns a list which repeats the previous signal, until a new 
    signal is given.

    Parameters
    ----------
    signals : list | pd.Series
        A series of signals. Zero values are treated as 'no signal'.

    Returns
    -------
    list
        A list of rolled signals.
    
    Examples
    --------
    >>> rolling_signal_list([0,1,0,0,0,-1,0,0,1,0,0])
        [0, 1, 1, 1, 1, -1, -1, -1, 1, 1, 1]
    
    """
    rolling_signals = [0]
    last_signal = rolling_signals[0]
    
    for i in range(1, len(signals)):
        if signals[i] != 0:
            last_signal = signals[i]
        
        rolling_signals.append(last_signal)
    
    return rolling_signals


def unroll_signal_list(signals: Union[list, pd.Series]) -> np.array:
    """Unrolls a rolled signal list.

    Parameters
    ----------
    signals : Union[list, pd.Series]
        DESCRIPTION.

    Returns
    -------
    unrolled_signals : np.array
        The unrolled signal series.
    
    See Also
    --------
    This function is the inverse of rolling_signal_list.
    
    Examples
    --------
    >>> unroll_signal_list([0, 1, 1, 1, 1, -1, -1, -1, 1, 1, 1])
        array([ 0.,  1.,  0.,  0.,  0., -1.,  0.,  0.,  1.,  0.,  0.])
        
    """
    new_list = np.zeros(len(signals))
    for i in range(len(signals)):
        if signals[i] != signals[i-1]:
            new_list[i] = signals[i]
    
    return new_list


def merge_signals(signal_1: list, signal_2: list) -> list:
    """Returns a single signal list which has merged two signal lists. 

    Parameters
    ----------
    signal_1 : list
        The first signal list.
    signal_2 : list
        The second signal list.

    Returns
    -------
    merged_signal_list : list
        The merged result of the two inputted signal series.
    
    Examples
    --------
    >>> s1 = [1,0,0,0,1,0]
    >>> s2 = [0,0,-1,0,0,-1]
    >>> merge_signals(s1, s2)
        [1, 0, -1, 0, 1, -1]
        
    """
    merged_signal_list = signal_1.copy()
    for i in range(len(signal_1)):
        if signal_2[i] != 0:
            merged_signal_list[i] = signal_2[i]
    
    return merged_signal_list


def build_grid_price_levels(grid_origin: float, grid_space: float, 
                            grid_levels: int, grid_price_space: float = None, 
                            pip_value: float = 0.0001) -> np.array:
    """Generates grid price levels.
    """
    # Calculate grid spacing in price units
    if grid_price_space is None:
        grid_price_space = grid_space*pip_value
    
    # Generate order_limit_price list 
    grid_price_levels = np.linspace(grid_origin - grid_levels*grid_price_space, 
                                     grid_origin + grid_levels*grid_price_space, 
                                     2*grid_levels + 1)
    
    return grid_price_levels


def build_grid(grid_origin: float, grid_space: float, grid_levels: int, 
               order_direction: int, order_type: str = 'stop-limit', 
               grid_price_space: float = None, pip_value: float = 0.0001, 
               take_distance: float = None, stop_distance: float = None, 
               stop_type: str = None) -> dict:
    """Generates a grid of orders.

    Parameters
    ----------
    grid_origin : float
        The origin of grid, specified as a price.
    grid_space : float
        The spacing between grid levels, specified as pip distance.
    grid_levels : int
        The number of grid levels either side of origin.
    order_direction : int
        The direction of each grid level order (1 for long, -1 for short).
    order_type : str, optional
        The order type of each grid level order. The default is 'stop-limit'.
    grid_price_space : float, optional
        The spacing between grid levels, specified as price units distance. 
        The default is None.
    pip_value : float, optional
        The instrument-specific pip value. The default is 0.0001.
    take_distance : float, optional
        The distance (in pips) of each order's take profit. The default is None.
    stop_distance : float, optional
        The distance (in pips) of each order's stop loss. The default is None.
    stop_type : str, optional
        The stop loss type. The default is None.

    Returns
    -------
    grid : dict
        A dictionary containing all orders on the grid.

    """
    
    # Check if stop_distance was provided without a stop_type
    if stop_distance is not None and stop_type is None:
        # set stop_type to 'limit' by default
        stop_type = 'limit'
    
    # Calculate grid spacing in price units
    if grid_price_space is None:
        grid_price_space = grid_space*pip_value
    
    # Generate order_limit_price list 
    order_limit_prices = np.linspace(grid_origin - grid_levels*grid_price_space, 
                                     grid_origin + grid_levels*grid_price_space, 
                                     2*grid_levels + 1)
    
    # Construct nominal order
    nominal_order = {}
    nominal_order["order_type"]         = order_type
    nominal_order["direction"]          = order_direction
    nominal_order["stop_distance"]      = stop_distance
    nominal_order["stop_type"]          = stop_type
    nominal_order["take_distance"]      = take_distance
    
    # Build grid
    grid = {}

    for order, limit_price in enumerate(order_limit_prices):
        grid[order] = nominal_order.copy()
        grid[order]["order_stop_price"]  = order_limit_prices[order]
        grid[order]["order_limit_price"] = order_limit_prices[order]
        
    return grid


def merge_grid_orders(grid_1: np.array, grid_2: np.array) -> np.array:
    """Merges grid dictionaries into one and re-labels order numbers so each
    order number is unique.
    """
    order_offset = len(grid_1)
    grid = grid_1.copy()
    
    for order_no in grid_2:
        grid[order_no + order_offset] = grid_2[order_no]
    
    return grid
    

def last_level_crossed(data: pd.DataFrame, base: float) -> list:
    """Returns a list containing the last grid level touched.
    The grid levels are determined by the base input variable, 
    which corresponds to the pip_space x pip_value.
    """
    last_level_crossed = np.nan
    levels_crossed = []
    for i in range(len(data)):
        high = data.High.values[i]
        low = data.Low.values[i]
        
        upper_prices = []
        lower_prices = []
        
        for price in [high, low]:    
            upper_prices.append(base*np.ceil(price/base))
            lower_prices.append(base*np.floor(price/base))
        
        if lower_prices[0] != lower_prices[1]:
            # Candle has crossed a level
            last_level_crossed = lower_prices[0]
        
        levels_crossed.append(last_level_crossed)
    
    return levels_crossed


def build_multiplier_grid(origin: float, direction: int, multiplier: float, 
                          no_levels: int, precision: int, spacing: float) -> list:
    """Constructs grid levels with a multiplying grid space.
    
    Parameters
    ----------
    origin : float 
        The origin of grid as price amount.
    
    direction : int
        The direction of grid (1 for long, -1 for short).
    
    multiplier : float
        The grid space multiplier when price moves away from the origin 
        opposite to direction.
    
    no_levels : int 
        The number of levels to calculate either side of the origin.
    
    precision : int 
        The instrument precision (eg. 4 for most currencies, 2 for JPY).
    
    spacing : float
        The spacing of the grid in price units.
    """
    
    levels = [i for i in range(1, no_levels + 1)]

    pos_levels = [round(origin + direction*spacing*i, precision) for i in levels]
    neg_spaces = [spacing*multiplier**(i) for i in levels]
    neg_levels = []
    prev_neg_level = origin
    for i in range(len(levels)):
        next_neg_level = prev_neg_level - direction*neg_spaces[i]
        prev_neg_level = next_neg_level
        neg_levels.append(round(next_neg_level, precision))
    
    grid = neg_levels + [origin] + pos_levels
    grid.sort()
    
    return grid


def last_level_touched(data: pd.DataFrame, grid: np.array) -> np.array:
    """Calculates the grid levels touched by price data.
    """
    # initialise with nan
    last_level_crossed = np.nan 
    
    levels_touched = []
    for i in range(len(data)):
        high = data.High.values[i]
        low = data.Low.values[i]
        
        upper_prices = []
        lower_prices = []
        
        for price in [high, low]:    
            # Calculate level above
            upper_prices.append(grid[next(x[0] for x in enumerate(grid) if x[1] > price)])
            
            # calculate level below
            first_level_below_index = next(x[0] for x in enumerate(grid[::-1]) if x[1] < price)
            lower_prices.append(grid[-(first_level_below_index+1)])
        
        if lower_prices[0] != lower_prices[1]:
            # Candle has crossed a level, since the level below the candle high
            # is different to the level below the candle low.
            # This essentially means the grid level is between candle low and high.
            last_level_crossed = lower_prices[0]
        
        levels_touched.append(last_level_crossed)
    
    return levels_touched


def stoch_rsi(data: pd.DataFrame, K_period: int = 3, D_period: int = 3, 
              RSI_length: int = 14, Stochastic_length: int = 14):
    """Stochastic RSI indicator.
    """
    rsi1 = TA.RSI(data, period=RSI_length)
    stoch = stochastic(rsi1, rsi1, rsi1, Stochastic_length)
    
    K = sma(stoch, K_period)
    D = sma(K, D_period)
    
    return K, D


def stochastic(high: pd.Series, low: pd.Series, close: pd.Series, 
               period: int = 14) -> pd.Series:
    """Stochastics indicator.
    """
    K = np.zeros(len(high))
    
    for i in range(period, len(high)):
        low_val     = min(low[i-period+1:i+1])
        high_val    = max(high[i-period+1:i+1])
        
        K[i]        = 100 * (close[i] - low_val)/(high_val - low_val)
        
    return K
    

def sma(data: pd.DataFrame, period: int = 14) -> list:
    """Smoothed Moving Average.
    """
    sma_list = []
    for i in range(len(data)):
        average = sum(data[i-period+1:i+1])/period
        sma_list.append(average)
    return sma_list


def ema(data: pd.DataFrame, period: int = 14, smoothing: int = 2) -> list:
    """Exponential Moving Average.
    """
    ema = [sum(data[:period]) / period]
    for price in data[period:]:
        ema.append((price * (smoothing / (1 + period))) + ema[-1] * (1 - (smoothing / (1 + period))))
    for i in range(period-1):
        ema.insert(0, np.nan)
    return ema


def true_range(data: pd.DataFrame, period: int = 14):
    """True range.
    """
    high_low = data['High'] - data['Low']
    high_close = np.abs(data['High'] - data['Close'].shift())
    low_close = np.abs(data['Low'] - data['Close'].shift())
    ranges = pd.concat([high_low, high_close, low_close], axis=1)
    true_range = np.max(ranges, axis=1)
    return true_range


def atr(data: pd.DataFrame, period: int = 14):
    """Average True Range.
    """
    tr = true_range(data, period)
    atr = tr.rolling(period).sum()/period
    return atr


def _conditional_ema(x, condition=1, n=14, s=2):
    'Conditional sampling EMA functtion'
    if type(condition) == int: condition = condition*np.ones(len(x))
    
    ema = np.zeros(len(x))
    for i in range(1, len(x)):
        if condition[i]:
            ema[i] = (x[i] - ema[i-1]) * (s/(1+n)) + ema[i-1]
        else:
            ema[i] = ema[i-1]
    
    return pd.Series(ema, x.index, name=f'{n} period conditional EMA')
    

def _conditional_sma(x, condition=1, n=14):
    'Conditional sampling SMA functtion'
    
    if type(condition) == int: condition = condition*np.ones(len(x))
    
    # Calculate SMA
    sma = x.rolling(n).mean()
    
    # Filter by condition
    sma = sma * condition
    
    return sma
    

def _stdev(x, n):
    'Standard deviation function'
    sd = np.sqrt(_conditional_sma(x**2, 1, n) - _conditional_sma(x, 1, n)**2)
    return sd
    

def _range_size(x, scale='AverageChange', qty=2.618, n=14):
    'Range size function'
    
    if scale == 'AverageChange':
        AC = _conditional_ema(abs(x - x.shift(1)), 1, n)
        rng_size = qty*AC
    elif scale == 'ATR':
        tr = TA.TR(x)
        atr = _conditional_ema(tr, 1, n)
        rng_size = qty*atr
    elif scale == 'StandardDeviation':
        sd = _stdev(x, n)
        rng_size = qty*sd
        
    return rng_size


def _calculate_range_filter(h, l, rng, n, rng_type, smooth, sn, av_rf, av_n):
    """Two type range filter function.
    """
    
    smoothed_range = _conditional_ema(rng, 1, sn)
    r = smoothed_range if smooth else rng
    r_filt = (h+l)/2
    
    if rng_type == 1:
        for i in range(1, len(h)):
            if h[i] - r[i] > r_filt[i-1]:
                r_filt[i] = h[i] - r[i]
            elif l[i] + r[i] < r_filt[i-1]:
                r_filt[i] = l[i] + r[i]
            else:
                r_filt[i] = r_filt[i-1]
    
    elif rng_type == 2:
        for i in range(1, len(h)):
            if h[i] >= r_filt[i-1] + r[i]:
                r_filt[i] = r_filt[i-1] + np.floor(abs(h[i] - r_filt[i-1])/r[i])*r[i]
            elif l[i] <= r_filt[i-1] - r[i]:
                r_filt[i] = r_filt[i-1] - np.floor(abs(l[i] - r_filt[i-1])/r[i])*r[i]
            else:
                r_filt[i] = r_filt[i-1]
    
    # Define nominal values
    r_filt1 = r_filt.copy()
    hi_band1 = r_filt1 + r
    lo_band1 = r_filt1 - r
    
    # Calculate indicator for averaged filter changes
    r_filt2 = _conditional_ema(r_filt1, r_filt1 != r_filt1.shift(1), av_n)
    hi_band2 = _conditional_ema(hi_band1, r_filt1 != r_filt1.shift(1), av_n)
    lo_band2 = _conditional_ema(lo_band1, r_filt1 != r_filt1.shift(1), av_n)
    
    # Assign indicator
    rng_filt = r_filt2 if av_rf else r_filt1
    hi_band = hi_band2 if av_rf else hi_band1
    lo_band = lo_band2 if av_rf else lo_band1
    
    # Construct output
    rfi = pd.DataFrame(data={'upper': hi_band, 'lower': lo_band, 
                             'rf': rng_filt}, index=rng_filt.index)
    
    # Classify filter direction
    rfi['fdir'] = np.sign(rfi.rf - rfi.rf.shift(1)).fillna(0)
    
    return rfi


import ib_insync
from datetime import datetime
from autotrader.brokers.trading import Order
from dateutil.relativedelta import relativedelta
from autotrader.brokers.broker_utils import BrokerUtils


class Utils(BrokerUtils):
    def __init__(self):
        pass
    
        
    def __repr__(self):
        return 'AutoTrader-InteractiveBrokers Utilities'
    
    
    def __str__(self):
        return 'AutoTrader-InteractiveBrokers Utilities'
    
    
    def check_response(self, response):
        """Checks API response for errors.

        Parameters
        ----------
        response : TYPE
            DESCRIPTION.

        Returns
        -------
        output : TYPE
            DESCRIPTION.
        """
        
        if response.status != 201:
            message = response.body['errorMessage']
        else:
            message = "Success."
            
        output = {'Status': response.status, 
                  'Message': message}
        
        return output
    
    
    def check_precision(self, pair, price):
        ''' Modify a price based on required ordering precision for pair. ''' 
        N               = self.get_precision(pair)
        corrected_price = round(price, N)
        
        return corrected_price
    
    
    def get_precision(self, pair):
        ''' Returns the allowable precision for a given pair '''
        
        response = self.api.account.instruments(accountID = self.ACCOUNT_ID, 
                                                instruments = pair)
        
        precision = response.body['instruments'][0].displayPrecision
        
        return precision
    
    
    def check_trade_size(self, pair, units):
        ''' Checks the requested trade size against the minimum trade size 
            allowed for the currency pair. '''
        response = self.api.account.instruments(accountID=self.ACCOUNT_ID, 
                                                instruments = pair)
        # minimum_units = response.body['instruments'][0].minimumTradeSize
        trade_unit_precision = response.body['instruments'][0].tradeUnitsPrecision
        
        return round(units, trade_unit_precision)
    
    
    @staticmethod
    def build_contract(order: Order) -> ib_insync.contract.Contract:
        """Builds IB contract from the order details.
        """
        instrument = order.instrument
        security_type = order.secType
        
        # Get contract object
        contract_object = getattr(ib_insync, security_type)
        
        if security_type == 'Stock':
            # symbol='', exchange='', currency=''
            exchange = order.exchange if order.exchange else 'SMART'
            currency = order.currency if order.currency else 'USD'
            contract = contract_object(symbol=instrument, exchange=exchange, currency=currency)
            
        elif security_type == 'Options':
            raise NotImplementedError(f"Contract building for {security_type.lower()} trading is not supported yet.")
            
        elif security_type == 'Future':
            # Requires order_details{'instrument', 'exchange', 'contract_month'}
            exchange = order.exchange if order.exchange else 'GLOBEX'
            currency = order.currency if order.currency else 'USD'
            contract_month = order.contract_month
            local_symbol = order.localSymbol if order.localSymbol else '' 
            contract = contract_object(symbol=instrument, 
                                       exchange=exchange, 
                                       currency=currency,
                                       lastTradeDateOrContractMonth=contract_month,
                                       localSymbol=local_symbol)
            
        elif security_type == 'ContFuture':
            raise NotImplementedError(f"Contract building for {security_type.lower()} trading is not supported yet.")
            
        elif security_type == 'Forex':
            # pair='', exchange='IDEALPRO', symbol='', currency='', **kwargs)
            exchange = order.exchange if order.exchange else 'IDEALPRO'
            contract = contract_object(pair=instrument, exchange=exchange)
            
        elif security_type == 'Index':
            raise NotImplementedError(f"Contract building for {security_type.lower()} trading is not supported yet.")
            
        elif security_type == 'CFD':
            # symbol='', exchange='', currency='',
            exchange = order.exchange if order.exchange else 'SMART'
            currency = order.currency if order.currency else 'USD'
            contract = contract_object(symbol=instrument, exchange=exchange, currency=currency)
            
        elif security_type == 'Commodity':
            raise NotImplementedError(f"Contract building for {security_type.lower()} trading is not supported yet.")
        elif security_type == 'Bond':
            raise NotImplementedError(f"Contract building for {security_type.lower()} trading is not supported yet.")
        elif security_type == 'FuturesOption':
            raise NotImplementedError(f"Contract building for {security_type.lower()} trading is not supported yet.")
        elif security_type == 'MutualFund':
            raise NotImplementedError(f"Contract building for {security_type.lower()} trading is not supported yet.")
        elif security_type == 'Warrant':
            raise NotImplementedError(f"Contract building for {security_type.lower()} trading is not supported yet.")
        elif security_type == 'Bag':
            raise NotImplementedError(f"Contract building for {security_type.lower()} trading is not supported yet.")
        elif security_type == 'Crypto':
            raise NotImplementedError(f"Contract building for {security_type.lower()} trading is not supported yet.")
        
        return contract
    
    
    @staticmethod
    def accsum_to_dict(account: str = None, data: list = None) -> dict:
        """Returns account summary list as a dictionary.

        Parameters
        ----------
        account : str
            DESCRIPTION.
        data : list
            DESCRIPTION.

        Returns
        -------
        out
            DESCRIPTION.
        """
        
        if account is None:
            account = 'All'
            
        out = {}
        for av in data:
            if av.account == account:
                out[av.tag] = {'value': av.value,
                               'currency': av.currency,
                               'modelCode': av.modelCode}
        
        return out
    
    @staticmethod
    def positionlist_to_dict(positions: list) -> dict:
        """Returns position list as a dictionary.

        Parameters
        ----------
        positions : list
            DESCRIPTION.

        Returns
        -------
        dict
            DESCRIPTION.

        """
        pass
    
    
    @staticmethod
    def _futures_expiry(dt: datetime = datetime.now(), months: int = 1) -> str:
        """Returns a string of format YYYYMM corresponding to the leading 
        contract month, from the inputted datetime object.

        Parameters
        ----------
        dt : datetime, optional
            The datetime object to convert to string. The default is datetime.now().
        months : int, optional
            The month offset from dt. The default is 1.

        Returns
        -------
        str
            A string corresponding to the expiry.
        """
        # TODO - if current date is past last trade date... error
        expiry_dt = dt + relativedelta(months=months)
        return expiry_dt.strftime('%Y') + expiry_dt.strftime('%m')
        
        


from __future__ import annotations
import numpy as np
from datetime import datetime
from autotrader.brokers.broker_utils import BrokerUtils


class Order:
    """AutoTrader Order object.

    Attributes
    ----------
    instrument : str
        The trading instrument of the order.
    direction : int
        The direction of the order (1 for long, -1 for short).
    order_type : str
        The type of order. The default is 'market'.
    size : float
        The number of units.
    stop_loss : float
        The price to set the stop-loss at.
    stop_distance : float
        The pip distance between the order price and the stop-loss.
    stop_type : str
        The type of stop-loss (limit or trailing). The default is 'limit'.
    take_profit : float
        The price to set the take-profit at.
    take_distance : float
        The pip distance between the order price and the take-profit.
    related_orders : list
        A list of related order/trade ID's.
    id : int
        The order ID.
    currency : str
        The base currency of the order (IB only).
    secType : str
        The security type (IB only).
    exchange : str
        The exchange on which the instrument is listed (IB only).
    contract_month : str
        The contract month string (IB only).
    localSymbol : str
        The exchange-specific instrument symbol (IB only).

    """
    def __init__(self, instrument: str = None, direction: int = None,
                 order_type: str = 'market', **kwargs) -> Order:
        
        # Required attributes
        self.instrument = instrument
        self.direction = direction
        self.order_type = order_type
        self.size = None
        self.order_price = None
        self.order_time = None
        self.order_limit_price = None
        self.order_stop_price = None
        
        self.HCF = 1
        
        self.stop_type = None
        self.stop_loss = None
        self.stop_distance = None
        
        self.take_profit = None
        self.take_distance = None
        
        self.related_orders = None
        
        self.data_name = None # When using custom DataStream
        
        # IB attributes
        self.currency = None
        self.secType = None
        self.exchange = None
        self.contract_month = None
        self.localSymbol = None
        
        # Oanda attributes
        self.trigger_price = "DEFAULT"
        
        self.reason = None
        
        self.strategy = None
        self.granularity = None
        self._sizing = None
        self._risk_pc = None
        
        # Meta-data
        self.id = None
        self.status = None # options: pending -> open -> cancelled | filled
        
        # Unpack kwargs
        for item in kwargs:
            setattr(self, item, kwargs[item])
        
        # Enforce stop type
        if self.stop_loss is not None:
            self.stop_type = self.stop_type if self.stop_type is \
                not None else 'limit'
    
    
    def __repr__(self):
        if self.size is not None:
            direction = 'long' if self.direction > 0 else 'short'
            return f'{round(self.size,3)} unit {direction} {self.instrument} {self.order_type} order'
        else:
            return self.__str__()
        
    
    def __str__(self):
        if self.instrument is None:
            return 'Blank order'
        else:
            return f'{self.instrument} {self.order_type} Order'
    
    
    def __call__(self, broker = None, order_price: float = None, 
                 order_time: datetime = datetime.now(), 
                 HCF: float = None) -> None:
        """Order object, called to initialise prior to submission to broker.

        Parameters
        ----------
        broker : TYPE, optional
            DESCRIPTION. The default is None.
        order_price : float, optional
            DESCRIPTION. The default is None.
        order_time : datetime, optional
            DESCRIPTION. The default is datetime.now().
        HCF : float, optional
            DESCRIPTION. The default is 1.

        Returns
        -------
        None
            DESCRIPTION.
        """
        self.order_price = order_price if order_price else self.order_price
        self.order_time = order_time if order_time else self.order_time
        self.HCF = HCF if HCF is not None else self.HCF
        
        # Enfore size scalar
        self.size = abs(self.size) if self.size is not None else self.size
        
        if self.order_type not in ['close']:
            self._set_working_price()
            self._calculate_exit_prices(broker)
            self._calculate_position_size(broker)
        
        self.status = 'submitted'
        self.submitted = True
        
        
    def _set_working_price(self, order_price: float = None) -> None:
        """Sets the Orders' working price, for calculating exit targets.

        Parameters
        ----------
        order_price : float, optional
            The order price.

        Returns
        -------
        None
            The working price will be saved as a class attribute.
        """
        order_price = order_price if order_price is not None \
            else self.order_price
        if self.order_type == 'limit' or self.order_type == 'stop-limit':
            self._working_price = self.order_limit_price
        else:
            self._working_price = order_price
        
    
    def _calculate_exit_prices(self, broker = None, 
                               working_price: float = None) -> None:
        """Calculates the prices of the exit targets from the pip distance
        values.

        Parameters
        ----------
        broker : AutoTrader Broker Interface, optional
            The autotrade-broker instance. The default is None.
        working_price : float, optional
            The working price used to calculate amount risked. The default is 
            None.

        Returns
        -------
        None
            The exit prices will be assigned to the order instance.
        """
        working_price = working_price if working_price is not None \
            else self._working_price
        
        if broker is None:
            utils = BrokerUtils()
        else:
            utils = broker.utils
        pip_value = utils.get_pip_ratio(self.instrument)

        # Calculate stop loss price
        if self.stop_loss is None and self.stop_distance is not None:
            # Stop loss provided as pip distance, convert to price
            self.stop_loss = working_price - np.sign(self.direction)*\
                self.stop_distance*pip_value
        
        # Calculate take profit price
        if self.take_profit is None and self.take_distance is not None:
            # Take profit pip distance specified, convert to price
            self.take_profit = working_price + np.sign(self.direction)*\
                self.take_distance*pip_value
        

    def _calculate_position_size(self, broker = None, working_price: float = None, 
                                 HCF: float = 1, risk_pc: float = 0,
                                 sizing: str | float = 'risk', 
                                 amount_risked: float = None) -> None:
        """Calculates trade size for order.

        Parameters
        ----------
        broker : AutoTrader Broker Interface, optional
            The autotrade-broker instance. The default is None.
        working_price : float, optional
            The working price used to calculate amount risked. The default is None.
        HCF : float, optional
            The home conversion factor. The default is 1.
        risk_pc : float, optional
            The percentage of the account NAV to risk on the trade. The default is 0.
        sizing : str | float, optional
            The sizing option. The default is 'risk'.
        amount_risked : float, optional
            The dollar amount risked on the trade. The default is None.

        Returns
        -------
        None
            The trade size will be assigned to the order instance.
        """ 
        working_price = working_price if working_price is not None \
            else self._working_price
        HCF = self.HCF if self.HCF is not None \
            else HCF
        sizing = self._sizing if self._sizing is not None \
            else sizing 
        risk_pc = self._risk_pc if self._risk_pc is not None \
            else risk_pc
        
        if self.size is None:
            amount_risked = amount_risked if amount_risked else \
                broker.get_NAV() * risk_pc / 100
            # Size not provided, need to calculate it
            if sizing == 'risk':
                size = broker.utils.get_size(self.instrument, amount_risked, 
                                             working_price, self.stop_loss, 
                                             HCF, self.stop_distance)
            else:
                self.size = sizing
            
            # Vectorise and save size
            self.size = size
    
    
    def _check_precision(self,):
        # TODO - implement
        raise NotImplementedError("This method has not been implemented yet.")
    
    
    def _validate(self,):
        # TODO - add order validation method, ie. for IB, check all attributes are
        # assigned (eg. sectype, etc)
        raise NotImplementedError("This method has not been implemented yet.")
    
    
    def as_dict(self) -> dict:
        """Converts Order object to dictionary.

        Returns
        -------
        dict
            The order instance returned as a dict object.

        Notes
        -----
        This method enables legacy code operation, returning order/trade
        objects as a dictionary.
        """
        return self.__dict__
    

    @classmethod
    def _from_dict(cls, order_dict: dict) -> Order:
        return Order(**order_dict)


class Trade(Order):
    """AutoTrader Trade.
    
    Attributes
    ----------
    unrealised_PL : float
        The floating PnL of the trade.
    margin_required : float
        The margin required to maintain the trade.
    time_filled : datetime
        The time at which the trade was filled.
    fill_price : float
        The price at which the trade was filled.
    last_price : float
        The last price observed for the instrument associated with the trade.
    last_time : datetime
        The last time observed for the instrument associated with the trade.
    exit_price : float
        The price at which the trade was closed.
    exit_time : datetime
        The time at which the trade was closed.
    fees : float
        The fees associated with the trade.
    parent_id : int 
        The ID of the order which spawned the trade.
    id : int
        The trade ID.
    status : str
        The status of the trade (open or closed).
    split : bool
        If the trade has been split.

    Notes
    -----
    When a trade is created from an Order, the Order will be marked as filled.
    """
    def __init__(self, order: Order = None, **kwargs) -> Trade:
        
        # Trade data
        self.unrealised_PL = 0
        self.margin_required = 0
        self.time_filled = None
        self.fill_price = None
        
        self.last_price = None
        self.last_time = None
        
        self.profit = 0
        self.balance = None
        self.exit_price = None
        self.exit_time = None
        self.fees = None
        
        # Meta data
        self.parent_id = None # ID of order which spawned trade
        self.id = None 
        self.status = None # options: open -> closed
        self.split = False
        
        # Unpack kwargs
        for item in kwargs:
            setattr(self, item, kwargs[item])
            
        # Inherit order attributes
        if order:
            self._inheret_order(order)
            order.status = 'filled'
            self.parent_id = order.id
        
    
    def __repr__(self):
        return f'{round(self.size,3)} unit {self.instrument} trade'
        
    
    def __str__(self):
        return 'AutoTrader Trade'
    
    
    def _inheret_order(self, order: Order) -> None:
        for attribute, value in order.__dict__.items():
            setattr(self, attribute, value)
            
    @classmethod
    def _split(cls, trade: Trade, split_units: float) -> Trade:
        """Splits parent trade into new trade object for partial trade 
        closures.
        
        split units are given to the new trade.
        """
        split_trade = cls()
        for attribute, value in trade.__dict__.items():
            setattr(split_trade, attribute, value)
        
        # Reset ID
        split_trade.parent_id = trade.parent_id
        split_trade.order_id = None
        
        # Transfer units
        split_trade.size = split_units
        trade.size -= split_units
        
        # Mark original trade as split
        trade.split = True
        
        return split_trade


class Position:
    """AutoTrader Position object.

    Attributes
    ----------
    instrument : str
        The trade instrument of the position.
    long_units : float
        The number of long units in the position.
    long_PL : float
        The PnL of the long units in the position.
    long_margin : float 
        The margin required to maintain the long units in the position.
    short_units : float 
        The number of short units in the position.
    short_PL : float 
        The PnL of the short units in the position.
    short_margin : float
        The margin required to maintain the short units in the position.
    total_margin : float 
        The total margin required to maintain the position.
    trade_IDs : list[int]
        The trade ID's associated with the position.
    net_position : float
        The total number of units in the position (IB only).
    PL : float
        The floating PnL (IB only).
    contracts : list
        The contracts associated with the position (IB only).
    portfolio_items : list
        The portfolio items associated with the position (IB only).
    """
    def __init__(self, **kwargs):
        self.instrument = None
        self.long_units = None
        self.long_PL = None
        self.long_margin = None
        self.short_units = None
        self.short_PL = None
        self.short_margin = None
        self.total_margin = None
        self.trade_IDs = None
        self.net_position = None
        
        # IB Attributes
        self.PL = None
        self.contracts = None
        self.portfolio_items = None
        
        for item in kwargs:
            setattr(self, item, kwargs[item])
    
    def __repr__(self):
        return f'AutoTrader Position in {self.instrument}'
        
    
    def __str__(self):
        return 'AutoTrader Position'
    
    
    def as_dict(self) -> dict:
        """Converts Position object to dictionary.

        Returns
        -------
        dict
            The Position instance returned as a dict object.

        Notes
        -----
        This method enables legacy code operation, returning order/trade
        objects as a dictionary.
        """
        return self.__dict__


if __name__ == '__main__':
    order_signal_dict = {'instrument': 'EUR_USD','order_type': 'market', 
                         'direction': 1, 'stop_loss': 1.22342}
    o = Order._from_dict(order_signal_dict)